# coding: utf-8
"""
Implementation of convert_to_schema(data, previous_schema, new_schema)
"""

import typing
from .generate_placeholder import generate_placeholder
from .utils import get_dimensionality_for_units
from ...models import ActionType
from .. import actions, errors, objects


def convert_to_schema(data: dict, previous_schema: dict, new_schema: dict) -> typing.Tuple[typing.Any, typing.Sequence[str]]:
    """
    Convert data from one schema to another.

    :param data: the sampledb object data
    :param previous_schema: the sampledb object schema for the given data
    :param new_schema: the target sampledb object schema
    :return: the converted data and a list of conversion warnings/notes
    """
    if new_schema == previous_schema and new_schema['type'] in ('bool', 'text', 'datetime', 'tags', 'sample', 'measurement', 'object_reference', 'quantity', 'array', 'objects', 'hazards'):
        return data, []

    if new_schema['type'] == 'tags' and previous_schema['type'] == 'text':
        tags = []
        for tag in data['text'].split(','):
            tag = tag.strip().lower()
            if tag not in tags:
                tags.append(tag)
        new_data = {
            '_type': 'tags',
            'tags': tags
        }
        return new_data, []
    if new_schema['type'] == 'object_reference' and previous_schema['type'] in ('sample', 'measurement'):
        action_type_compatible = False
        if 'action_type_id' in new_schema and new_schema['action_type_id'] is not None:
            if previous_schema['type'] == 'sample' and new_schema['action_type_id'] == ActionType.SAMPLE_CREATION:
                action_type_compatible = True
            if previous_schema['type'] == 'measurement' and new_schema['action_type_id'] == ActionType.MEASUREMENT:
                action_type_compatible = True
        else:
            action_type_compatible = True
        action_compatible = False
        if 'action_id' in new_schema and new_schema['action_id'] is not None:
            if 'object_id' in data:
                try:
                    referenced_object = objects.get_object(data['object_id'])
                    action_id = referenced_object.action_id
                except errors.ObjectDoesNotExistError:
                    pass
                else:
                    if action_id == new_schema['action_id']:
                        action_compatible = True
            else:
                action_compatible = True
        else:
            action_compatible = True
        if action_type_compatible and action_compatible:
            return {
                '_type': 'object_reference',
                'object_id': data['object_id']
            }, []
    if previous_schema['type'] == 'object_reference' and new_schema['type'] in ('sample', 'measurement'):
        if 'action_type_id' in previous_schema:
            if new_schema['type'] == 'sample' and previous_schema['action_type_id'] == ActionType.SAMPLE_CREATION:
                return data, []
            if new_schema['type'] == 'measurement' and previous_schema['action_type_id'] == ActionType.MEASUREMENT:
                return data, []
        if 'action_type_id' not in previous_schema or previous_schema['action_type_id'] is None:
            if 'object_id' in data:
                try:
                    referenced_object = objects.get_object(data['object_id'])
                    action = actions.get_action(referenced_object.action_id)
                    action_type_id = action.type_id
                except errors.ObjectDoesNotExistError:
                    pass
                except errors.ActionDoesNotExistError:
                    pass
                else:
                    if new_schema['type'] == 'sample' and action_type_id == ActionType.SAMPLE_CREATION:
                        return data, []
                    if new_schema['type'] == 'measurement' and action_type_id == ActionType.MEASUREMENT:
                        return data, []
    if previous_schema['type'] != new_schema['type']:
        return generate_placeholder(new_schema), ["Unable to convert property '{}' from type '{}' to type '{}'.".format(new_schema['title'], previous_schema['type'], new_schema['type'])]
    if new_schema['type'] in ('bool', 'text', 'datetime', 'tags', 'sample', 'measurement', 'hazards', 'user'):
        return data, []
    if new_schema['type'] == 'object_reference':
        referenced_object = None
        action_type_compatible = 'action_type_id' not in new_schema or new_schema['action_type_id'] is None or ('action_type_id' in previous_schema and new_schema['action_type_id'] == previous_schema['action_type_id'])
        if not action_type_compatible and ('action_type_id' not in previous_schema or previous_schema['action_type_id'] is None):
            if 'object_id' in data:
                try:
                    referenced_object = objects.get_object(data['object_id'])
                    action = actions.get_action(referenced_object.action_id)
                    action_type_id = action.type_id
                except errors.ObjectDoesNotExistError:
                    pass
                except errors.ActionDoesNotExistError:
                    pass
                else:
                    if action_type_id == new_schema['action_type_id']:
                        action_type_compatible = True
            else:
                action_type_compatible = True
        action_compatible = 'action_id' not in new_schema or new_schema['action_id'] is None or ('action_id' in previous_schema and new_schema['action_id'] == previous_schema['action_id'])
        if not action_compatible and ('action_id' not in previous_schema or previous_schema['action_id'] is None):
            if 'object_id' in data:
                try:
                    if referenced_object is None:
                        referenced_object = objects.get_object(data['object_id'])
                    action_id = referenced_object.action_id
                except errors.ObjectDoesNotExistError:
                    pass
                else:
                    if action_id == new_schema['action_id']:
                        action_compatible = True
            else:
                action_compatible = True
        if action_type_compatible and action_compatible:
            return data, []
    if new_schema['type'] == 'quantity':
        previous_dimensionality = get_dimensionality_for_units(previous_schema['units'])
        new_dimensionality = get_dimensionality_for_units(new_schema['units'])
        if new_dimensionality == previous_dimensionality:
            return data, []
        return generate_placeholder(new_schema), ["Unable to convert quantity '{}' to different dimensionality: {} -> {}".format(new_schema['title'], previous_dimensionality, new_dimensionality)]
    if new_schema['type'] == 'object':
        upgrade_warnings = []
        new_data = generate_placeholder(new_schema)
        for property_name, property_value in data.items():
            if property_name in new_schema['properties']:
                new_property_value, property_upgrade_warnings = convert_to_schema(property_value, previous_schema['properties'][property_name], new_schema['properties'][property_name])
                if new_property_value is not None:
                    new_data[property_name] = new_property_value
                for upgrade_warning in property_upgrade_warnings:
                    if upgrade_warning not in upgrade_warnings:
                        upgrade_warnings.append(upgrade_warning)
        return new_data, upgrade_warnings
    if new_schema['type'] == 'array':
        new_data = []
        upgrade_warnings = []
        for item in data:
            new_item, item_upgrade_warnings = convert_to_schema(item, previous_schema['items'], new_schema['items'])
            new_data.append(new_item)
            for upgrade_warning in item_upgrade_warnings:
                if upgrade_warning not in upgrade_warnings:
                    upgrade_warnings.append(upgrade_warning)
        return new_data, upgrade_warnings
    return generate_placeholder(new_schema), ["Unable to convert property '{}' of type '{}'.".format(new_schema['title'], new_schema['type'])]



# coding: utf-8
"""

"""

from .. import db
from .users import User
from .actions import Action
from .versioned_json_object_tables import VersionedJSONSerializableObjectTables

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


class Object(VersionedJSONSerializableObjectTables.VersionedJSONSerializableObject):
    pass


Objects = VersionedJSONSerializableObjectTables(
    'objects',
    object_type=Object,
    user_id_column=User.id,
    action_id_column=Action.id,
    action_schema_column=Action.schema,
    metadata=db.metadata
)



# coding: utf-8
"""

"""

import flask

from . import frontend
from ..logic.errors import UserIsReadonlyError

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


@frontend.app_errorhandler(400)
def bad_request(error):
    try:
        return flask.render_template('errors/400.html'), 400
    except Exception:
        return 'Bad Request', 400


@frontend.app_errorhandler(403)
def forbidden(error):
    try:
        return flask.render_template('errors/403.html'), 403
    except Exception:
        return 'Forbidden', 403


@frontend.app_errorhandler(404)
def file_not_found(error):
    try:
        return flask.render_template('errors/404.html'), 404
    except Exception:
        return 'Not found', 404


@frontend.app_errorhandler(500)
def internal_server_error(error):
    try:
        return flask.render_template('errors/500.html'), 500
    except Exception:
        return 'Internal Server Error', 500


@frontend.errorhandler(UserIsReadonlyError)
def user_is_readonly_error(error):
    return flask.render_template('errors/user_is_readonly.html'), 403



# coding: utf-8
"""
Replace the type column by the type_id column in the actions table.
"""

import os

from ..actions import ActionType

MIGRATION_INDEX = 31
MIGRATION_NAME, _ = os.path.splitext(os.path.basename(__file__))


def run(db):
    # Skip migration by condition
    column_names = db.session.execute("""
        SELECT column_name
        FROM information_schema.columns
        WHERE table_name = 'actions'
    """).fetchall()
    if ('type_id',) in column_names:
        return False

    # Perform migration
    db.session.execute("""
        ALTER TABLE actions
        ADD type_id INTEGER NULL
    """)

    default_action_types = [
        {
            'type_id': ActionType.SAMPLE_CREATION,
            'type': 'SAMPLE_CREATION'
        },
        {
            'type_id': ActionType.MEASUREMENT,
            'type': 'MEASUREMENT'
        },
        {
            'type_id': ActionType.SIMULATION,
            'type': 'SIMULATION',
        }
    ]
    for action_type in default_action_types:
        db.session.execute("""
            UPDATE actions
            SET type_id = :type_id
            WHERE type::text = :type
        """, params=action_type)

    db.session.execute("""
        ALTER TABLE actions
        ALTER type_id SET NOT NULL
    """)

    db.session.execute("""
        ALTER TABLE actions
        ADD CONSTRAINT fk_actions_type_id FOREIGN KEY (type_id) REFERENCES action_types (id)
    """)

    db.session.execute("""
        ALTER TABLE actions
        DROP type
    """)
    return True



# coding: utf-8
"""

"""

from .. import db
from . import Objects


class ObjectPublication(db.Model):
    __tablename__ = 'object_publications'

    object_id = db.Column(db.Integer, db.ForeignKey(Objects.object_id_column), nullable=False)
    doi = db.Column(db.String, nullable=False)
    title = db.Column(db.String)
    object_name = db.Column(db.String)

    __table_args__ = (
        db.PrimaryKeyConstraint(object_id, doi),
        {},
    )



# coding: utf-8
"""

"""

import enum

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


class Permissions(enum.Enum):
    NONE = 0
    READ = 1
    WRITE = 2  # includes READ
    GRANT = 3  # includes READ and WRITE

    def __contains__(self, item: 'Permissions') -> bool:
        return self.value >= item.value

    def __str__(self):
        return self.name.lower()

    def __le__(self, other):
        return self.value <= other.value

    def __lt__(self, other):
        return self.value < other.value

    @staticmethod
    def from_name(name):
        members = {
            'none': Permissions.NONE,
            'read': Permissions.READ,
            'write': Permissions.WRITE,
            'grant': Permissions.GRANT,
        }
        try:
            return members[name.lower()]
        except KeyError:
            raise ValueError('Invalid name')



# coding: utf-8
"""
Logic module for user managed groups

Users can create uniquely named groups and use them to more easily manage
permissions for samples or measurements. At this time, all group members
are equally able to add new members to a group and to modify a group's
name and description.

This module provides a procedural interface to the underlying database model,
with the immutable Group class accessible to the developer instead of an
SQLAlchemy ORM class.

As the group models use flask-sqlalchemy however, the functions in this
module should be called from within a Flask application context.
"""
import collections
import datetime
import typing

import flask
from sqlalchemy.exc import IntegrityError

from .. import db
from ..models import groups
from .users import get_user
from .security_tokens import generate_token
from .notifications import create_notification_for_being_invited_to_a_group
from . import errors


# group names are limited to this (semi-arbitrary) length
MAX_GROUP_NAME_LENGTH = 100


class Group(collections.namedtuple('Group', ['id', 'name', 'description'])):
    """
    This class provides an immutable wrapper around models.groups.Group.
    """

    def __new__(cls, id: int, name: str, description: str):
        self = super(Group, cls).__new__(cls, id, name, description)
        return self

    @classmethod
    def from_database(cls, group: groups.Group) -> 'Group':
        return Group(id=group.id, name=group.name, description=group.description)


class GroupInvitation(collections.namedtuple('GroupInvitation', ['id', 'group_id', 'user_id', 'inviter_id', 'utc_datetime', 'accepted'])):
    """
    This class provides an immutable wrapper around models.groups.GroupInvitation.
    """

    def __new__(cls, id: int, group_id: int, user_id: int, inviter_id: int, utc_datetime: datetime.datetime, accepted: bool):
        self = super(GroupInvitation, cls).__new__(cls, id, group_id, user_id, inviter_id, utc_datetime, accepted)
        return self

    @classmethod
    def from_database(cls, group_invitation: groups.GroupInvitation) -> 'GroupInvitation':
        return GroupInvitation(
            id=group_invitation.id,
            group_id=group_invitation.group_id,
            user_id=group_invitation.user_id,
            inviter_id=group_invitation.inviter_id,
            utc_datetime=group_invitation.utc_datetime,
            accepted=group_invitation.accepted
        )

    @property
    def expired(self):
        expiration_datetime = self.utc_datetime + datetime.timedelta(seconds=flask.current_app.config['INVITATION_TIME_LIMIT'])
        return datetime.datetime.utcnow() >= expiration_datetime


def create_group(name: str, description: str, initial_user_id: int) -> Group:
    """
    Creates a new group with the given name and description and adds an
    initial user to it.

    :param name: the unique name of the group
    :param description: a (possibly empty) description for the group
    :param initial_user_id: the user ID of the initial group member
    :return: the newly created group
    :raise errors.InvalidGroupNameError: when the group name is empty or more
        than MAX_GROUP_NAME_LENGTH characters long
    :raise errors.UserDoesNotExistError: when no user with the given
        user ID exists
    :raise errors.GroupAlreadyExistsError: when another group with the given
        name already exists
    """
    if not 1 <= len(name) <= MAX_GROUP_NAME_LENGTH:
        raise errors.InvalidGroupNameError()
    user = get_user(initial_user_id)
    group = groups.Group(name=name, description=description)
    group.members.append(user)
    db.session.add(group)
    try:
        db.session.commit()
    except IntegrityError:
        db.session.rollback()
        existing_group = groups.Group.query.filter_by(name=name).first()
        if existing_group is not None:
            raise errors.GroupAlreadyExistsError()
        raise
    return group


def update_group(group_id: int, name: str, description: str = '') -> None:
    """
    Updates the group's name and description.

    :param group_id: the ID of an existing group
    :param name: the new unique group name
    :param description: the new group description
    :raise errors.InvalidGroupNameError: when the group name is empty or more
        than MAX_GROUP_NAME_LENGTH characters long
    :raise errors.GroupDoesNotExistError: when no group with the given
        group ID exists
    :raise errors.GroupAlreadyExistsError: when another group with the given
        name already exists
    """
    if not 1 <= len(name) <= MAX_GROUP_NAME_LENGTH:
        raise errors.InvalidGroupNameError()
    group = groups.Group.query.get(group_id)
    if group is None:
        raise errors.GroupDoesNotExistError()
    group.name = name
    group.description = description
    db.session.add(group)
    try:
        db.session.commit()
    except IntegrityError:
        db.session.rollback()
        existing_group = groups.Group.query.filter_by(name=name).first()
        if existing_group is not None:
            raise errors.GroupAlreadyExistsError()
        raise


def delete_group(group_id: int) -> None:
    """
    Deletes the given group, freeing up the name and removing all permissions
    that might have been granted to the group.

    This is also implicitly done when the last member leaves a group.

    :param group_id: the ID of an existing group
    :raise errors.GroupDoesNotExistError: when no group with the given
        group ID exists
    """
    group = groups.Group.query.get(group_id)
    if group is None:
        raise errors.GroupDoesNotExistError()
    # group permissions and group default permissions will be deleted due to
    # ondelete = "CASCADE" in the model. No need to delete them manually here.
    db.session.delete(group)
    db.session.commit()


def get_group(group_id: int) -> Group:
    """
    Returns the group with the given ID.

    :param group_id: the ID of an existing group
    :return: the group
    :raise errors.GroupDoesNotExistError: when no group with the given
        group ID exists
    """
    group = groups.Group.query.get(group_id)
    if group is None:
        raise errors.GroupDoesNotExistError()
    return Group.from_database(group)


def get_groups() -> typing.List[Group]:
    """
    Returns a list of all existing groups.

    :return: the list of groups
    """
    return [Group.from_database(group) for group in groups.Group.query.all()]


def get_group_member_ids(group_id: int) -> typing.List[int]:
    """
    Returns a list of the user IDs of all members of the group with the given
    group ID.

    :param group_id: the ID of an existing group
    :return: the member ID list
    :raise errors.GroupDoesNotExistError: when no group with the given
        group ID exists
    """
    group = groups.Group.query.get(group_id)
    if group is None:
        raise errors.GroupDoesNotExistError()
    return [user.id for user in group.members]


def get_user_groups(user_id: int) -> typing.List[Group]:
    """
    Returns a list of the group IDs of all groups the user with the given
    user ID is a member of.

    :param group_id: the ID of an existing user
    :return: the member ID list
    :raise errors.UserDoesNotExistError: when no user with the given
        user ID exists
    """
    user = get_user(user_id)
    return [Group.from_database(group) for group in user.groups]


def invite_user_to_group(group_id: int, user_id: int, inviter_id: int) -> None:
    """
    Invite a user to a group.

    :param group_id: the ID of an existing group
    :param user_id: the ID of an existing user
    :param inviter_id: the ID of who invited this user to the group
    :raise errors.GroupDoesNotExistError: when no group with the given group
        ID exists
    :raise errors.UserDoesNotExistError: when no user with the given user ID
        or inviter ID exists
    :raise errors.UserAlreadyMemberOfGroupError: when the user with the given
        user ID already is a member of this group
    """
    # ensure the inviter exists
    get_user(inviter_id)
    group = groups.Group.query.get(group_id)
    if group is None:
        raise errors.GroupDoesNotExistError()
    user = get_user(user_id)
    if user in group.members:
        raise errors.UserAlreadyMemberOfGroupError()

    invitation = groups.GroupInvitation(
        group_id=group_id,
        user_id=user_id,
        inviter_id=inviter_id,
        utc_datetime=datetime.datetime.utcnow()
    )
    db.session.add(invitation)
    db.session.commit()
    token = generate_token(
        {
            'invitation_id': invitation.id,
            'user_id': user.id,
            'group_id': group_id
        },
        salt='invite_to_group',
        secret_key=flask.current_app.config['SECRET_KEY']
    )
    expiration_time_limit = flask.current_app.config['INVITATION_TIME_LIMIT']
    expiration_utc_datetime = datetime.datetime.utcnow() + datetime.timedelta(seconds=expiration_time_limit)
    confirmation_url = flask.url_for("frontend.group", group_id=group_id, token=token, _external=True)
    create_notification_for_being_invited_to_a_group(user_id, group_id, inviter_id, confirmation_url, expiration_utc_datetime)


def add_user_to_group(group_id: int, user_id: int) -> None:
    """
    Adds the user with the given user ID to the group with the given group ID.

    :param group_id: the ID of an existing group
    :param user_id: the ID of an existing user
    :raise errors.GroupDoesNotExistError: when no group with the given
        group ID exists
    :raise errors.UserDoesNotExistError: when no user with the given
        user ID exists
    :raise errors.UserAlreadyMemberOfGroupError: when the user is already
        a member of the group
    """
    group = groups.Group.query.get(group_id)
    if group is None:
        raise errors.GroupDoesNotExistError()
    user = get_user(user_id)
    if user in group.members:
        raise errors.UserAlreadyMemberOfGroupError()
    group.members.append(user)
    invitations = get_group_invitations(
        group_id=group_id,
        user_id=user_id,
        include_expired_invitations=False,
        include_accepted_invitations=False
    )
    for invitation in invitations:
        invitation = groups.GroupInvitation.query.filter_by(id=invitation.id).first()
        invitation.accepted = True
        db.session.add(invitation)
    db.session.commit()


def remove_user_from_group(group_id: int, user_id: int) -> None:
    """
    Removes the user with the given user ID from the group with the given
    group ID.

    If there are no members left in the group, it is deleted and its name is
    freed up to be used again for a new group.

    :param group_id: the ID of an existing group
    :param user_id: the ID of an existing user
    :raise errors.GroupDoesNotExistError: when no group with the given
        group ID exists
    :raise errors.UserDoesNotExistError: when no user with the given
        user ID exists
    :raise errors.UserNotMemberOfGroupError: when the user is not a member of
        the group
    """
    group = groups.Group.query.get(group_id)
    if group is None:
        raise errors.GroupDoesNotExistError()
    user = get_user(user_id)
    if user not in group.members:
        raise errors.UserNotMemberOfGroupError()
    group.members.remove(user)
    if not group.members:
        db.session.delete(group)
    db.session.commit()


def get_group_invitations(
        group_id: int,
        user_id: typing.Optional[int] = None,
        include_accepted_invitations: bool = True,
        include_expired_invitations: bool = True
) -> typing.List[GroupInvitation]:
    """
    Get all (current) invitations for a user to join a group.

    :param group_id: the ID of an existing group
    :param user_id: the ID of an existing user or None
    :param include_accepted_invitations: whether or not to include invitations
        that have already been accepted by the user
    :param include_expired_invitations: whether or not to include invitations
        that have already expired
    :return: the list of group invitations
    :raise errors.GroupDoesNotExistError: when no group with the given
        group ID exists
    """
    group_invitations_query = groups.GroupInvitation.query.filter_by(group_id=group_id)
    if user_id is not None:
        group_invitations_query = group_invitations_query.filter_by(user_id=user_id)
    if not include_accepted_invitations:
        group_invitations_query = group_invitations_query.filter_by(accepted=False)
    if not include_expired_invitations:
        expiration_time_limit = flask.current_app.config['INVITATION_TIME_LIMIT']
        expired_invitation_datetime = datetime.datetime.utcnow() - datetime.timedelta(seconds=expiration_time_limit)
        group_invitations_query = group_invitations_query.filter(groups.GroupInvitation.utc_datetime > expired_invitation_datetime)
    group_invitations = group_invitations_query.order_by(groups.GroupInvitation.utc_datetime).all()
    if not group_invitations:
        # ensure that a group with this ID exists
        get_group(group_id)
    return [
        GroupInvitation.from_database(group_invitation)
        for group_invitation in group_invitations
    ]


def get_group_invitation(invitation_id: int) -> GroupInvitation:
    """
    Get an existing invitation.

    :param invitation_id: the ID of an existing invitation
    :return: the invitation
    :raise errors.GroupInvitationDoesNotExistError: when no invitation with
        the given ID exists
    """
    invitation = groups.GroupInvitation.query.filter_by(id=invitation_id).first()
    if invitation is None:
        raise errors.GroupInvitationDoesNotExistError()
    return GroupInvitation.from_database(invitation)



# coding: utf-8
"""
RESTful API for SampleDB
"""

import flask
from flask_restful import Resource

from .authentication import object_permissions_required, Permissions
from ...logic import users, groups, projects, errors, object_permissions

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


class UserObjectPermissions(Resource):
    @object_permissions_required(Permissions.READ)
    def get(self, object_id: int, user_id: int):
        include_instrument_responsible_users = 'include_instrument_responsible_users' in flask.request.args
        include_groups = 'include_groups' in flask.request.args
        include_projects = 'include_projects' in flask.request.args
        permissions = object_permissions.get_object_permissions_for_users(
            object_id=object_id,
            include_instrument_responsible_users=include_instrument_responsible_users,
            include_groups=include_groups,
            include_projects=include_projects
        ).get(user_id, Permissions.NONE)
        if permissions == Permissions.NONE:
            try:
                users.get_user(user_id)
            except errors.UserDoesNotExistError:
                return {
                    "message": "user {} does not exist".format(user_id)
                }, 404
        return permissions.name.lower()

    @object_permissions_required(Permissions.GRANT)
    def put(self, object_id: int, user_id: int):
        try:
            users.get_user(user_id)
        except errors.UserDoesNotExistError:
            return {
                "message": "user {} does not exist".format(user_id)
            }, 404
        request_json = flask.request.get_json(force=True)
        if not isinstance(request_json, str):
            return {
                "message": "JSON string body required"
            }, 400
        try:
            permissions = Permissions.from_name(request_json)
        except ValueError:
            return {
                "message": "Permissions name required"
            }, 400
        object_permissions.set_user_object_permissions(object_id, user_id, permissions)
        return permissions.name.lower(), 200


class UsersObjectPermissions(Resource):
    @object_permissions_required(Permissions.READ)
    def get(self, object_id: int):
        include_instrument_responsible_users = 'include_instrument_responsible_users' in flask.request.args
        include_groups = 'include_groups' in flask.request.args
        include_projects = 'include_projects' in flask.request.args
        permissions = object_permissions.get_object_permissions_for_users(
            object_id=object_id,
            include_instrument_responsible_users=include_instrument_responsible_users,
            include_groups=include_groups,
            include_projects=include_projects
        )
        return {
            user_id: permissions.name.lower()
            for user_id, permissions in permissions.items()
        }, 200


class GroupObjectPermissions(Resource):
    @object_permissions_required(Permissions.READ)
    def get(self, object_id: int, group_id: int):
        include_projects = 'include_projects' in flask.request.args
        permissions = object_permissions.get_object_permissions_for_groups(
            object_id=object_id,
            include_projects=include_projects
        ).get(group_id, Permissions.NONE)
        if permissions == Permissions.NONE:
            try:
                groups.get_group(group_id)
            except errors.GroupDoesNotExistError:
                return {
                    "message": "group {} does not exist".format(group_id)
                }, 404
        return permissions.name.lower()

    @object_permissions_required(Permissions.GRANT)
    def put(self, object_id: int, group_id: int):
        try:
            groups.get_group(group_id)
        except errors.GroupDoesNotExistError:
            return {
                "message": "group {} does not exist".format(group_id)
            }, 404
        request_json = flask.request.get_json(force=True)
        if not isinstance(request_json, str):
            return {
                "message": "JSON string body required"
            }, 400
        try:
            permissions = Permissions.from_name(request_json)
        except ValueError:
            return {
                "message": "Permissions name required"
            }, 400
        object_permissions.set_group_object_permissions(object_id, group_id, permissions)
        return permissions.name.lower(), 200


class GroupsObjectPermissions(Resource):
    @object_permissions_required(Permissions.READ)
    def get(self, object_id: int):
        include_projects = 'include_projects' in flask.request.args
        permissions = object_permissions.get_object_permissions_for_groups(
            object_id=object_id,
            include_projects=include_projects
        )
        return {
            group_id: permissions.name.lower()
            for group_id, permissions in permissions.items()
        }, 200


class ProjectObjectPermissions(Resource):
    @object_permissions_required(Permissions.READ)
    def get(self, object_id: int, project_id: int):
        permissions = object_permissions.get_object_permissions_for_projects(
            object_id=object_id
        ).get(project_id, Permissions.NONE)
        if permissions == Permissions.NONE:
            try:
                projects.get_project(project_id)
            except errors.ProjectDoesNotExistError:
                return {
                    "message": "project {} does not exist".format(project_id)
                }, 404
        return permissions.name.lower()

    @object_permissions_required(Permissions.GRANT)
    def put(self, object_id: int, project_id: int):
        try:
            projects.get_project(project_id)
        except errors.ProjectDoesNotExistError:
            return {
                "message": "project {} does not exist".format(project_id)
            }, 404
        request_json = flask.request.get_json(force=True)
        if not isinstance(request_json, str):
            return {
                "message": "JSON string body required"
            }, 400
        try:
            permissions = Permissions.from_name(request_json)
        except ValueError:
            return {
                "message": "Permissions name required"
            }, 400
        object_permissions.set_project_object_permissions(object_id, project_id, permissions)
        return permissions.name.lower(), 200


class ProjectsObjectPermissions(Resource):
    @object_permissions_required(Permissions.READ)
    def get(self, object_id: int):
        permissions = object_permissions.get_object_permissions_for_projects(
            object_id=object_id,
        )
        return {
            project_id: permissions.name.lower()
            for project_id, permissions in permissions.items()
        }, 200


class PublicObjectPermissions(Resource):
    @object_permissions_required(Permissions.READ)
    def get(self, object_id: int):
        is_public = object_permissions.object_is_public(
            object_id=object_id,
        )
        return is_public, 200

    @object_permissions_required(Permissions.GRANT)
    def put(self, object_id: int):
        request_json = flask.request.get_json(force=True)
        if not isinstance(request_json, bool):
            return {
                "message": "JSON boolean body required"
            }, 400
        is_public = bool(request_json)
        object_permissions.set_object_public(object_id, is_public)
        return is_public, 200



# coding: utf-8
"""

"""

from flask_wtf import FlaskForm
from wtforms import FieldList, FormField, SelectField, IntegerField, TextAreaField, HiddenField, FileField, StringField
from wtforms.validators import InputRequired, ValidationError, url

from ..logic.object_permissions import Permissions
from ..logic.publications import simplify_doi
from ..logic.errors import InvalidDOIError

from .validators import ObjectIdValidator


class ObjectUserPermissionsForm(FlaskForm):
    user_id = IntegerField(
        validators=[InputRequired()]
    )
    permissions = SelectField(
        choices=[(p.name.lower(), p.name.lower()) for p in (Permissions.NONE, Permissions.READ, Permissions.WRITE, Permissions.GRANT)],
        validators=[InputRequired()]
    )


class ObjectGroupPermissionsForm(FlaskForm):
    group_id = IntegerField(
        validators=[InputRequired()]
    )
    permissions = SelectField(
        choices=[(p.name.lower(), p.name.lower()) for p in (Permissions.NONE, Permissions.READ, Permissions.WRITE, Permissions.GRANT)],
        validators=[InputRequired()]
    )


class ObjectProjectPermissionsForm(FlaskForm):
    project_id = IntegerField(
        validators=[InputRequired()]
    )
    permissions = SelectField(
        choices=[(p.name.lower(), p.name.lower()) for p in (Permissions.NONE, Permissions.READ, Permissions.WRITE, Permissions.GRANT)],
        validators=[InputRequired()]
    )


class ObjectPermissionsForm(FlaskForm):
    public_permissions = SelectField(
        choices=[(p.name.lower(), p.name.lower()) for p in (Permissions.NONE, Permissions.READ)],
        validators=[InputRequired()]
    )
    user_permissions = FieldList(FormField(ObjectUserPermissionsForm), min_entries=0)
    group_permissions = FieldList(FormField(ObjectGroupPermissionsForm), min_entries=0)
    project_permissions = FieldList(FormField(ObjectProjectPermissionsForm), min_entries=0)


class ObjectForm(FlaskForm):
    pass


class ObjectVersionRestoreForm(FlaskForm):
    pass


class CommentForm(FlaskForm):
    content = TextAreaField(validators=[InputRequired()])


class FileForm(FlaskForm):
    file_source = HiddenField(validators=[InputRequired()])
    file_names = HiddenField()
    local_files = FileField()

    def validate_file_source(form, field):
        if field.data not in ['local']:
            raise ValidationError('Invalid file source')


class ExternalLinkForm(FlaskForm):
    url = StringField(validators=[url()])


class FileInformationForm(FlaskForm):
    title = StringField()
    description = TextAreaField()


class FileHidingForm(FlaskForm):
    reason = TextAreaField()


class ObjectLocationAssignmentForm(FlaskForm):
    location = SelectField(validators=[InputRequired()])
    responsible_user = SelectField(validators=[InputRequired()])
    description = StringField()


class ObjectPublicationForm(FlaskForm):
    doi = StringField()
    title = StringField()
    object_name = StringField()

    def validate_doi(form, field):
        try:
            field.data = simplify_doi(field.data)
        except InvalidDOIError:
            raise ValidationError('Please enter a valid DOI')


class CopyPermissionsForm(FlaskForm):
    object_id = SelectField(validators=[ObjectIdValidator(Permissions.GRANT), InputRequired()], validate_choice=False)



# coding: utf-8
"""
Logic module for user managed projects

Users can create uniquely named projects and use them to more easily manage
permissions for samples or measurements. Users with GRANT permissions for a
project can add new users and manage other users' permissions.

When a project is given permissions for an object, users will get whatever
level of permissions is lower: the project's object permissions or the user's
project permissions. Same is true for a group.

This module provides a procedural interface to the underlying database model,
with the immutable Project class accessible to the developer instead of an
SQLAlchemy ORM class.

As the project models use flask-sqlalchemy however, the functions in this
module should be called from within a Flask application context.
"""

from sqlalchemy.exc import IntegrityError
import collections
import datetime
import typing
import flask

from .. import db
from ..models import projects, Permissions, UserProjectPermissions, GroupProjectPermissions, SubprojectRelationship
from .users import get_user
from .security_tokens import generate_token
from . import groups
from . import errors
from .notifications import create_notification_for_being_invited_to_a_project


# project names are limited to this (semi-arbitrary) length
MAX_PROJECT_NAME_LENGTH = 100


class Project(collections.namedtuple('Project', ['id', 'name', 'description'])):
    """
    This class provides an immutable wrapper around models.projects.Project.
    """

    def __new__(cls, id: int, name: str, description: str):
        self = super(Project, cls).__new__(cls, id, name, description)
        return self

    @classmethod
    def from_database(cls, project: projects.Project) -> 'Project':
        return Project(id=project.id, name=project.name, description=project.description)


class ProjectInvitation(collections.namedtuple('ProjectInvitation', ['id', 'project_id', 'user_id', 'inviter_id', 'utc_datetime', 'accepted'])):
    """
    This class provides an immutable wrapper around models.projects.ProjectInvitation.
    """

    def __new__(cls, id: int, project_id: int, user_id: int, inviter_id: int, utc_datetime: datetime.datetime, accepted: bool):
        self = super(ProjectInvitation, cls).__new__(cls, id, project_id, user_id, inviter_id, utc_datetime, accepted)
        return self

    @classmethod
    def from_database(cls, project_invitation: projects.ProjectInvitation) -> 'ProjectInvitation':
        return ProjectInvitation(
            id=project_invitation.id,
            project_id=project_invitation.project_id,
            user_id=project_invitation.user_id,
            inviter_id=project_invitation.inviter_id,
            utc_datetime=project_invitation.utc_datetime,
            accepted=project_invitation.accepted
        )

    @property
    def expired(self):
        expiration_datetime = self.utc_datetime + datetime.timedelta(seconds=flask.current_app.config['INVITATION_TIME_LIMIT'])
        return datetime.datetime.utcnow() >= expiration_datetime


def create_project(name: str, description: str, initial_user_id: int) -> Project:
    """
    Creates a new project with the given name and description and adds an
    initial user to it.

    :param name: the unique name of the project
    :param description: a (possibly empty) description for the project
    :param initial_user_id: the user ID of the initial project member
    :return: the newly created project
    :raise errors.InvalidProjectNameError: when the project name is empty or more
        than MAX_PROJECT_NAME_LENGTH characters long
    :raise errors.UserDoesNotExistError: when no user with the given
        user ID exists
    :raise errors.ProjectAlreadyExistsError: when another project with the given
        name already exists
    """
    if not 1 <= len(name) <= MAX_PROJECT_NAME_LENGTH:
        raise errors.InvalidProjectNameError()
    user = get_user(initial_user_id)
    project = projects.Project(name=name, description=description)
    db.session.add(project)
    try:
        db.session.commit()
    except IntegrityError:
        db.session.rollback()
        existing_project = projects.Project.query.filter_by(name=name).first()
        if existing_project is not None:
            raise errors.ProjectAlreadyExistsError()
        raise
    user_project_permissions = projects.UserProjectPermissions(project_id=project.id, user_id=user.id, permissions=Permissions.GRANT)
    db.session.add(user_project_permissions)
    db.session.commit()
    return Project.from_database(project)


def update_project(project_id: int, name: str, description: str = '') -> None:
    """
    Updates the project's name and description.

    :param project_id: the ID of an existing project
    :param name: the new unique project name
    :param description: the new project description
    :raise errors.InvalidProjectNameError: when the project name is empty or more
        than MAX_PROJECT_NAME_LENGTH characters long
    :raise errors.ProjectDoesNotExistError: when no project with the given
        project ID exists
    :raise errors.ProjectAlreadyExistsError: when another project with the given
        name already exists
    """
    if not 1 <= len(name) <= MAX_PROJECT_NAME_LENGTH:
        raise errors.InvalidProjectNameError()
    project = projects.Project.query.get(project_id)
    if project is None:
        raise errors.ProjectDoesNotExistError()
    project.name = name
    project.description = description
    db.session.add(project)
    try:
        db.session.commit()
    except IntegrityError:
        db.session.rollback()
        existing_project = projects.Project.query.filter_by(name=name).first()
        if existing_project is not None:
            raise errors.ProjectAlreadyExistsError()
        raise


def delete_project(project_id: int) -> None:
    """
    Deletes the given project, freeing up the name and removing all permissions
    that might have been granted to the project.

    This is also implicitly done when the last member leaves a project.

    :param project_id: the ID of an existing project
    :raise errors.ProjectDoesNotExistError: when no project with the given
        project ID exists
    """
    project = projects.Project.query.get(project_id)
    if project is None:
        raise errors.ProjectDoesNotExistError()
    # project object permissions and project default permissions will be
    # deleted due to ondelete = "CASCADE" in the model. No need to delete
    # them manually here.
    db.session.delete(project)
    db.session.commit()


def get_project(project_id: int) -> Project:
    """
    Returns the project with the given ID.

    :param project_id: the ID of an existing project
    :return: the project
    :raise errors.ProjectDoesNotExistError: when no project with the given
        project ID exists
    """
    project = projects.Project.query.get(project_id)
    if project is None:
        raise errors.ProjectDoesNotExistError()
    return Project.from_database(project)


def get_projects() -> typing.List[Project]:
    """
    Returns a list of all existing projects.

    :return: the list of projects
    """
    return [Project.from_database(project) for project in projects.Project.query.all()]


def get_project_member_user_ids_and_permissions(project_id: int, include_groups: bool = False) -> typing.Dict[int, Permissions]:
    """
    Returns a dict of the user IDs of all members of the project with the
    given project ID, mapping them to their respective permissions.

    :param project_id: the ID of an existing project
    :param include_groups: whether or not groups membership should be
        considered as well
    :return: the member ID to member permission dict
    :raise errors.ProjectDoesNotExistError: when no project with the given
        project ID exists
    """
    project = projects.Project.query.get(project_id)
    if project is None:
        raise errors.ProjectDoesNotExistError()
    user_permissions = projects.UserProjectPermissions.query.filter_by(project_id=project_id).all()
    member_user_ids_and_permissions = {
        single_user_permissions.user_id: single_user_permissions.permissions
        for single_user_permissions in user_permissions
    }
    if include_groups:
        member_group_ids_and_permissions = get_project_member_group_ids_and_permissions(project_id)
        for member_group_id, permissions in member_group_ids_and_permissions.items():
            group_member_ids = groups.get_group_member_ids(member_group_id)
            for user_id in group_member_ids:
                if user_id not in member_user_ids_and_permissions or permissions not in member_user_ids_and_permissions[user_id]:
                    member_user_ids_and_permissions[user_id] = permissions
    return member_user_ids_and_permissions


def get_user_project_permissions(project_id: int, user_id: int, include_groups: bool = False) -> Permissions:
    """
    Returns a the permissions of a single user for a project, optionally
    including the users group memberships.

    :param project_id: the ID of an existing project
    :param user_id: the ID of an existing user
    :param include_groups: whether or not groups membership should be
        considered as well
    :return: permissions of this user
    :raise errors.ProjectDoesNotExistError: when no project with the given
        project ID exists
    :raise errors.UserDoesNotExistError: when no user with the given user ID
        exists
    """
    user_permissions = projects.UserProjectPermissions.query.filter_by(project_id=project_id, user_id=user_id).first()
    if user_permissions:
        permissions = user_permissions.permissions
    else:
        # verify that project exists or raise error
        get_project(project_id)
        permissions = Permissions.NONE
    if include_groups:
        group_permissions = projects.GroupProjectPermissions.query.filter_by(project_id=project_id).all()
        for group_permission in group_permissions:
            if user_id in groups.get_group_member_ids(group_permission.group_id) and permissions in group_permission.permissions:
                permissions = group_permission.permissions
    return permissions


def get_project_member_group_ids_and_permissions(project_id: int) -> typing.Dict[int, Permissions]:
    """
    Returns a dict of the group IDs of all members of the project with the
    given project ID, mapping them to their respective permissions.

    :param project_id: the ID of an existing project
    :return: the member ID to member permission dict
    :raise errors.ProjectDoesNotExistError: when no project with the given
        project ID exists
    """
    project = projects.Project.query.get(project_id)
    if project is None:
        raise errors.ProjectDoesNotExistError()
    group_permissions = projects.GroupProjectPermissions.query.filter_by(project_id=project_id).all()
    return {
        single_group_permissions.group_id: single_group_permissions.permissions
        for single_group_permissions in group_permissions
    }


def get_user_projects(user_id: int, include_groups: bool = False) -> typing.List[Project]:
    """
    Returns a list of the project IDs of all projects the user with the given
    user ID is a member of.

    :param project_id: the ID of an existing project
    :param include_groups: whether or not groups membership should be
        considered as well
    :return: the member ID list
    :raise errors.UserDoesNotExistError: when no user with the given
        user ID exists
    """
    user = get_user(user_id)
    user_permissions = projects.UserProjectPermissions.query.filter_by(user_id=user.id).all()
    project_ids = {
        user_permission.project_id for user_permission in user_permissions
    }
    if include_groups:
        user_groups = groups.get_user_groups(user_id)
        for group in user_groups:
            group_permissions = projects.GroupProjectPermissions.query.filter_by(group_id=group.id).all()
            for group_permission in group_permissions:
                project_ids.add(group_permission.project_id)
    return [get_project(project_id) for project_id in project_ids]


def invite_user_to_project(project_id: int, user_id: int, inviter_id: int, add_to_parent_project_ids: typing.Sequence[int] = ()) -> None:
    """
    Sends an invitation mail for a project to a user.

    :param project_id: the ID of an existing project
    :param user_id:  the ID of an existing users
    :param inviter_id: the ID of who invited this user to the project
    :param add_to_parent_project_ids: list of IDs of parent projects to which
        the user should also be added
    :raise errors.ProjectDoesNotExistError: when no project with the given
        project ID exists
    :raise errors.UserDoesNotExistError: when no user with the given user ID
        or inviter ID exists
    :raise errors.UserAlreadyMemberOfProjectError: when the user with the given
        user ID already is a member of this project
    """
    project = projects.Project.query.get(project_id)
    if project is None:
        raise errors.ProjectDoesNotExistError()
    get_user(user_id)
    if Permissions.READ in get_user_project_permissions(project_id, user_id):
        raise errors.UserAlreadyMemberOfProjectError()
    invitation = projects.ProjectInvitation(
        project_id=project_id,
        user_id=user_id,
        inviter_id=inviter_id,
        utc_datetime=datetime.datetime.utcnow()
    )
    db.session.add(invitation)
    db.session.commit()
    token = generate_token(
        {
            'invitation_id': invitation.id,
            'user_id': user_id,
            'project_id': project_id,
            'other_project_ids': add_to_parent_project_ids
        },
        salt='invite_to_project',
        secret_key=flask.current_app.config['SECRET_KEY']
    )
    expiration_time_limit = flask.current_app.config['INVITATION_TIME_LIMIT']
    expiration_utc_datetime = datetime.datetime.utcnow() + datetime.timedelta(seconds=expiration_time_limit)
    confirmation_url = flask.url_for("frontend.project", project_id=project_id, token=token, _external=True)
    create_notification_for_being_invited_to_a_project(user_id, project_id, inviter_id, confirmation_url, expiration_utc_datetime)


def add_user_to_project(project_id: int, user_id: int, permissions: Permissions, other_project_ids: typing.Sequence[int] = ()) -> None:
    """
    Adds the user with the given user ID to the project with the given project ID.

    :param project_id: the ID of an existing project
    :param user_id: the ID of an existing user
    :param permissions: the permissions of the user for this project
    :param other_project_ids: IDs of (parent) projects this user should also
        be added to
    :raise errors.ProjectDoesNotExistError: when no project with the given
        project ID exists
    :raise errors.UserDoesNotExistError: when no user with the given
        user ID exists
    :raise errors.UserAlreadyMemberOfProjectError: when the user is already
        a member of the project
    """
    if permissions == permissions.NONE:
        # project members with no permissions are not stored
        return
    project = projects.Project.query.get(project_id)
    if project is None:
        raise errors.ProjectDoesNotExistError()
    user = get_user(user_id)
    existing_permissions = projects.UserProjectPermissions.query.filter_by(
        project_id=project_id,
        user_id=user.id
    ).first()
    if existing_permissions is not None:
        raise errors.UserAlreadyMemberOfProjectError()
    user_permissions = UserProjectPermissions(project_id=project_id, user_id=user_id, permissions=permissions)
    db.session.add(user_permissions)
    invitations = get_project_invitations(
        project_id=project_id,
        user_id=user_id,
        include_expired_invitations=False,
        include_accepted_invitations=False
    )
    for invitation in invitations:
        invitation = projects.ProjectInvitation.query.filter_by(id=invitation.id).first()
        invitation.accepted = True
        db.session.add(invitation)
    db.session.commit()
    if other_project_ids:
        ancestor_project_ids = get_ancestor_project_ids(project_id, only_if_child_can_add_users_to_ancestor=True)
        for other_project_id in other_project_ids:
            if other_project_id in ancestor_project_ids:
                try:
                    add_user_to_project(other_project_id, user_id, permissions)
                except errors.UserAlreadyMemberOfProjectError:
                    pass


def add_group_to_project(project_id: int, group_id: int, permissions: Permissions) -> None:
    """
    Adds the group with the given group ID to the project with the given project ID.

    :param project_id: the ID of an existing project
    :param group_id: the ID of an existing group
    :param permissions: the permissions for the group
    :raise errors.ProjectDoesNotExistError: when no project with the given
        project ID exists
    :raise errors.GroupDoesNotExistError: when no group with the given
        group ID exists
    :raise errors.GroupAlreadyMemberOfProjectError: when the group is already
        a member of the project
    """
    if permissions == permissions.NONE:
        # project members with no permissions are not stored
        return
    project = projects.Project.query.get(project_id)
    if project is None:
        raise errors.ProjectDoesNotExistError()
    group = groups.get_group(group_id)
    existing_permissions = projects.GroupProjectPermissions.query.filter_by(
        project_id=project_id,
        group_id=group.id
    ).first()
    if existing_permissions is not None:
        raise errors.GroupAlreadyMemberOfProjectError()
    group_permissions = GroupProjectPermissions(project_id=project_id, group_id=group_id, permissions=permissions)
    db.session.add(group_permissions)
    db.session.commit()


def remove_user_from_project(project_id: int, user_id: int) -> None:
    """
    Removes the user with the given user ID from the project with the given
    project ID.

    If there are no member users or groups left in the project, it is deleted
    and its name is freed up to be used again for a new project.

    :param project_id: the ID of an existing project
    :param user_id: the ID of an existing user
    :raise errors.ProjectDoesNotExistError: when no project with the given
        project ID exists
    :raise errors.UserDoesNotExistError: when no user with the given
        user ID exists
    :raise errors.UserNotMemberOfProjectError: when the user is not a member of
        the project
    :raise errors.NoMemberWithGrantPermissionsForProjectError: when there would
        be no user or group with grant permissions left
    """
    project = projects.Project.query.get(project_id)
    if project is None:
        raise errors.ProjectDoesNotExistError()
    user = get_user(user_id)
    existing_permissions = projects.UserProjectPermissions.query.filter_by(
        project_id=project_id,
        user_id=user.id
    ).first()
    if existing_permissions is None:
        raise errors.UserNotMemberOfProjectError()
    other_group_permissions = projects.GroupProjectPermissions.query.filter_by(project_id=project_id).all()
    other_user_permissions = projects.UserProjectPermissions.query.filter_by(project_id=project_id).filter(UserProjectPermissions.user_id != user_id).all()
    if existing_permissions.permissions == Permissions.GRANT and (other_user_permissions or other_group_permissions):
        any_other_user_with_grant = any(user_permissions.permissions == Permissions.GRANT for user_permissions in other_user_permissions)
        if not any_other_user_with_grant:
            raise errors.NoMemberWithGrantPermissionsForProjectError()
    if not other_user_permissions and not other_group_permissions:
        db.session.delete(project)
    else:
        db.session.delete(existing_permissions)
    db.session.commit()


def remove_group_from_project(project_id: int, group_id: int) -> None:
    """
    Removes the group with the given group ID from the project with the given
    project ID.

    :param project_id: the ID of an existing project
    :param group_id: the ID of an existing group
    :raise errors.ProjectDoesNotExistError: when no project with the given
        project ID exists
    :raise errors.GroupDoesNotExistError: when no group with the given
        group ID exists
    :raise errors.GroupNotMemberOfProjectError: when the group is not a member of
        the project
    """
    project = projects.Project.query.get(project_id)
    if project is None:
        raise errors.ProjectDoesNotExistError()
    group = groups.get_group(group_id)
    existing_permissions = projects.GroupProjectPermissions.query.filter_by(
        project_id=project_id,
        group_id=group.id
    ).first()
    if existing_permissions is None:
        raise errors.GroupNotMemberOfProjectError()
    db.session.delete(existing_permissions)
    db.session.commit()


def update_user_project_permissions(project_id: int, user_id: int, permissions: Permissions) -> None:
    """
    Updates the permissions of the user with the given user ID for the project
    with the given project ID.

    Removes the user from the project if permissions is NONE.

    :param project_id: the ID of an existing project
    :param user_id: the ID of an existing user
    :param permissions: the new permissions
    :raise errors.ProjectDoesNotExistError: when no project with the given
        project ID exists
    :raise errors.UserDoesNotExistError: when no user with the given
        user ID exists
    :raise errors.UserNotMemberOfProjectError: when the user is not a member of
        the project
    :raise errors.NoMemberWithGrantPermissionsForProjectError: when there would
        be no user or group with grant permissions left
    """

    if permissions == Permissions.NONE:
        remove_user_from_project(project_id=project_id, user_id=user_id)
        return
    project = projects.Project.query.get(project_id)
    if project is None:
        raise errors.ProjectDoesNotExistError()
    get_user(user_id)
    existing_permissions = projects.UserProjectPermissions.query.filter_by(
        project_id=project_id,
        user_id=user_id
    ).first()
    if existing_permissions is None:
        raise errors.UserNotMemberOfProjectError()
    if existing_permissions.permissions == permissions:
        return

    if existing_permissions.permissions == Permissions.GRANT:
        other_user_permissions = projects.UserProjectPermissions.query.filter_by(project_id=project_id).filter(UserProjectPermissions.user_id != user_id).all()
        any_other_user_with_grant = any(user_permissions.permissions == Permissions.GRANT for user_permissions in other_user_permissions)
        if not any_other_user_with_grant:
            raise errors.NoMemberWithGrantPermissionsForProjectError()

    existing_permissions.permissions = permissions
    db.session.add(existing_permissions)
    db.session.commit()


def update_group_project_permissions(project_id: int, group_id: int, permissions: Permissions) -> None:
    """
    Updates the permissions of the group with the given group ID for the project
    with the given project ID.

    Removes the group from the project if permissions is NONE.

    :param project_id: the ID of an existing project
    :param group_id: the ID of an existing group
    :param permissions: the new permissions
    :raise errors.ProjectDoesNotExistError: when no project with the given
        project ID exists
    :raise errors.GroupDoesNotExistError: when no group with the given
        group ID exists
    :raise errors.GroupNotMemberOfProjectError: when the group is not a member of
        the project
    """

    if permissions == Permissions.NONE:
        remove_group_from_project(project_id=project_id, group_id=group_id)
        return
    project = projects.Project.query.get(project_id)
    if project is None:
        raise errors.ProjectDoesNotExistError()
    groups.get_group(group_id)
    existing_permissions = projects.GroupProjectPermissions.query.filter_by(
        project_id=project_id,
        group_id=group_id
    ).first()
    if existing_permissions is None:
        raise errors.GroupNotMemberOfProjectError()

    existing_permissions.permissions = permissions
    db.session.add(existing_permissions)
    db.session.commit()


def get_group_projects(group_id: int) -> typing.List[Project]:
    """
    Returns a list of the project IDs of all projects the group with the given
    group ID is a member of.

    :param project_id: the ID of an existing project
    :return: the member ID list
    :raise errors.GroupDoesNotExistError: when no group with the given
        group ID exists
    """
    group = groups.get_group(group_id)
    group_permissions = projects.GroupProjectPermissions.query.filter_by(group_id=group.id).all()
    project_ids = {
        group_permission.project_id for group_permission in group_permissions
    }
    return [get_project(project_id) for project_id in project_ids]


def filter_child_project_candidates(parent_project_id: int, child_project_ids: typing.List[int]) -> typing.List[int]:
    """
    Filters a list of project IDs to find which of them may become child
    projects of another given project.

    :param parent_project_id: the ID of an existing project
    :param child_project_ids: a list of IDs of existing projects
    :return: child_project_ids without IDs of projects which may not become
        child projects of parent_project_id
    """
    # Prevent relationship of a project with itself
    for child_project_id in child_project_ids[:]:
        if child_project_id == parent_project_id:
            child_project_ids.remove(child_project_id)
    # Prevent relationship duplication
    for child_project_id in child_project_ids[:]:
        if SubprojectRelationship.query.filter_by(parent_project_id=parent_project_id, child_project_id=child_project_id).first():
            child_project_ids.remove(child_project_id)
    # Prevent cycles
    if child_project_ids:
        ancestor_project_ids = get_ancestor_project_ids(parent_project_id)
        for child_project_id in child_project_ids[:]:
            if child_project_id in ancestor_project_ids:
                child_project_ids.remove(child_project_id)
                continue
    return child_project_ids


def create_subproject_relationship(parent_project_id: int, child_project_id: int, child_can_add_users_to_parent: bool = False) -> None:
    """
    Creates a subproject relationship between two existing projects.

    :param parent_project_id: the ID of an existing project
    :param child_project_id: the ID of an existing project
    :param child_can_add_users_to_parent: whether or not to allow users with
        GRANT permission on the child project to add users to the parent
    :raise errors.InvalidSubprojectRelationshipError: if the relationship
        would cause a cyclic relationship or exceed the maximum subproject
        depth
    """
    if not filter_child_project_candidates(parent_project_id, [child_project_id]):
        raise errors.InvalidSubprojectRelationshipError()
    subproject_relationship = SubprojectRelationship(parent_project_id=parent_project_id, child_project_id=child_project_id, child_can_add_users_to_parent=child_can_add_users_to_parent)
    db.session.add(subproject_relationship)
    db.session.commit()


def delete_subproject_relationship(parent_project_id: int, child_project_id: int) -> None:
    """
    Deletes an existing subproject relationship between two existing projects.

    :param parent_project_id: the ID of an existing project
    :param child_project_id: the ID of an existing project
    :raise errors.SubprojectRelationshipDoesNotExistError: if the relationship
        does not exist
    """
    subproject_relationship = SubprojectRelationship.query.filter_by(
        parent_project_id=parent_project_id,
        child_project_id=child_project_id
    ).first()
    if subproject_relationship is None:
        raise errors.SubprojectRelationshipDoesNotExistError()
    db.session.delete(subproject_relationship)
    db.session.commit()


def get_parent_project_ids(project_id: int, only_if_child_can_add_users_to_parent: bool = False) -> typing.List[int]:
    """
    Return the list of parent project IDs for an existing project.

    :param project_id: the ID of an existing project
    :param only_if_child_can_add_users_to_parent: whether or not to only show
        those parent projects, which someone with GRANT permissions on this
        project can add users to (transitively)
    :return: list of project IDs
    """
    subproject_relationships: typing.Iterable[SubprojectRelationship] = SubprojectRelationship.query.filter_by(
        child_project_id=project_id
    ).all()
    parent_project_ids = []
    for subproject_relationship in subproject_relationships:
        if subproject_relationship.child_can_add_users_to_parent or not only_if_child_can_add_users_to_parent:
            parent_project_ids.append(subproject_relationship.parent_project_id)
    return parent_project_ids


def get_child_project_ids(project_id: int) -> typing.List[int]:
    """
    Return the list of child project IDs for an existing project.

    :param project_id: the ID of an existing project
    :return: list of project IDs
    """
    subproject_relationships = SubprojectRelationship.query.filter_by(
        parent_project_id=project_id
    ).all()
    child_project_ids = []
    for subproject_relationship in subproject_relationships:
        child_project_ids.append(subproject_relationship.child_project_id)
    return child_project_ids


def get_ancestor_project_ids(project_id: int, only_if_child_can_add_users_to_ancestor: bool = False) -> typing.Set[int]:
    """
    Return the list of (transitive) ancestor project IDs for an existing
    project.

    :param project_id: the ID of an existing project
    :param only_if_child_can_add_users_to_ancestor: whether or not to only show
        those ancestor projects, which someone with GRANT permissions on this
        project can add users to (transitively)
    :return: set of project IDs
    """
    ancestor_project_ids = set()
    project_id_queue = [project_id]
    while project_id_queue:
        ancestor_project_id = project_id_queue.pop(0)
        if ancestor_project_id in ancestor_project_ids:
            continue
        if ancestor_project_id != project_id:
            ancestor_project_ids.add(ancestor_project_id)
        project_id_queue.extend(get_parent_project_ids(ancestor_project_id, only_if_child_can_add_users_to_parent=only_if_child_can_add_users_to_ancestor))
    return ancestor_project_ids


def get_descendent_project_ids(project_id: int) -> typing.Set[int]:
    """
    Return the list of (transitive) descendent project IDs for an existing
    project.

    :param project_id: the ID of an existing project
    :return: set of project IDs
    """
    descendent_project_ids = set()
    project_id_queue = [project_id]
    while project_id_queue:
        descendent_project_id = project_id_queue.pop(0)
        if descendent_project_id in descendent_project_ids:
            continue
        if descendent_project_id != project_id:
            descendent_project_ids.add(descendent_project_id)
        project_id_queue.extend(get_child_project_ids(descendent_project_id))
    return descendent_project_ids


def can_child_add_users_to_parent_project(child_project_id: int, parent_project_id: int) -> bool:
    """
    Return whether or not the subproject relationship allows adding users to
    the parent project.

    :param child_project_id: the ID of an existing child project
    :param parent_project_id: the ID of an existing parent project
    :return: whether or not someone with GRANT permissions on the child
        project can add users to the parent project
    """
    subproject_relationship: typing.Optional[SubprojectRelationship] = SubprojectRelationship.query.filter_by(
        child_project_id=child_project_id,
        parent_project_id=parent_project_id
    ).first()
    if subproject_relationship is None:
        return False
    return subproject_relationship.child_can_add_users_to_parent


def get_project_invitations(
        project_id: int,
        user_id: typing.Optional[int] = None,
        include_accepted_invitations: bool = True,
        include_expired_invitations: bool = True
) -> typing.List[ProjectInvitation]:
    """
    Get all (current) invitations for a user to join a project.

    :param project_id: the ID of an existing project
    :param user_id: the ID of an existing user or None
    :param include_accepted_invitations: whether or not to include invitations
        that have already been accepted by the user
    :param include_expired_invitations: whether or not to include invitations
        that have already expired
    :return: the list of project invitations
    :raise errors.ProjectDoesNotExistError: when no project with the given
        project ID exists
    """
    project_invitations_query = projects.ProjectInvitation.query.filter_by(project_id=project_id)
    if user_id is not None:
        project_invitations_query = project_invitations_query.filter_by(user_id=user_id)
    if not include_accepted_invitations:
        project_invitations_query = project_invitations_query.filter_by(accepted=False)
    if not include_expired_invitations:
        expiration_time_limit = flask.current_app.config['INVITATION_TIME_LIMIT']
        expired_invitation_datetime = datetime.datetime.utcnow() - datetime.timedelta(seconds=expiration_time_limit)
        project_invitations_query.filter(projects.ProjectInvitation.utc_datetime > expired_invitation_datetime)
    project_invitations = project_invitations_query.order_by(projects.ProjectInvitation.utc_datetime).all()
    if not project_invitations:
        # ensure that a project with this ID exists
        get_project(project_id)
    return [
        ProjectInvitation.from_database(project_invitation)
        for project_invitation in project_invitations
    ]


def get_project_invitation(invitation_id: int) -> ProjectInvitation:
    """
    Get an existing invitation.

    :param invitation_id: the ID of an existing invitation
    :return: the invitation
    :raise errors.ProjectInvitationDoesNotExistError: when no invitation with
        the given ID exists
    """
    invitation = projects.ProjectInvitation.query.filter_by(id=invitation_id).first()
    if invitation is None:
        raise errors.ProjectInvitationDoesNotExistError()
    return ProjectInvitation.from_database(invitation)


def get_project_id_hierarchy_list(project_ids: typing.List[int]) -> typing.List[typing.Tuple[int, int]]:
    """
    Return a list of project IDs and the degree that they are related to a root project.

    This list is sorted so that child projects follow their parents. If a
    project is the child of multiple parent projects, it and its descendents
    will be included repeatedly.

    :param project_ids: the project IDs to include
    :return: a list of tuples containing the project level and ID
    """
    child_project_ids = {
        project_id: [
            child_project_id
            for child_project_id in get_child_project_ids(project_id)
            if child_project_id in project_ids
        ]
        for project_id in project_ids
    }

    root_project_ids = [
        project_id
        for project_id in project_ids
        if not any(
            project_id in ids
            for ids in child_project_ids.values()
        )
    ]

    project_id_hierarchy_list = []
    parent_project_queue = [
        (0, project_id)
        for project_id in sorted(root_project_ids, reverse=True)
    ]
    while parent_project_queue:
        level, parent_project_id = parent_project_queue.pop()
        project_id_hierarchy_list.append((level, parent_project_id))
        parent_project_queue.extend([
            (level + 1, project_id)
            for project_id in reversed(child_project_ids[parent_project_id])
        ])
        if not parent_project_queue:
            # In case of cyclical relationships, which should be
            # impossible, projects may not be included yet, so
            # if there are any left, this picks the one with the lowest ID.
            remaining_project_ids = set(project_ids) - {
                project_id
                for level, project_id in project_id_hierarchy_list
            }
            if remaining_project_ids:
                parent_project_queue.append((0, sorted(remaining_project_ids)[0]))

    return project_id_hierarchy_list



# coding: utf-8
"""
Script for updating an instrument in SampleDB.

Usage: python -m sampledb update_instrument <instrument_id> <name> <description>
"""

import sys
from .. import create_app
from ..logic.instruments import update_instrument, get_instrument
from ..logic.errors import InstrumentDoesNotExistError


def main(arguments):
    if len(arguments) != 3:
        print(__doc__)
        exit(1)
    instrument_id, name, description = arguments
    try:
        instrument_id = int(instrument_id)
    except ValueError:
        print("Error: instrument_id must be an integer", file=sys.stderr)
        exit(1)
    app = create_app()
    with app.app_context():
        try:
            get_instrument(instrument_id)
        except InstrumentDoesNotExistError:
            print('Error: no instrument with this id exists', file=sys.stderr)
            exit(1)
        update_instrument(instrument_id, name, description)
        print("Success: the instrument has been updated in SampleDB")



# coding: utf-8
"""
Add users_can_create_log_entries column to instruments table.
"""

import os

MIGRATION_INDEX = 20
MIGRATION_NAME, _ = os.path.splitext(os.path.basename(__file__))


def run(db):
    # Skip migration by condition
    column_names = db.session.execute("""
        SELECT column_name
        FROM information_schema.columns
        WHERE table_name = 'instruments'
    """).fetchall()
    if ('users_can_create_log_entries',) in column_names:
        return False

    # Perform migration
    db.session.execute("""
        ALTER TABLE instruments
        ADD users_can_create_log_entries BOOLEAN NOT NULL DEFAULT FALSE
    """)
    return True



# coding: utf-8
"""
Script for creating a user of type OTHER in SampleDB.

Usage: python -m sampledb create_other_user <name> <email>
"""

import os
import sys
from .. import create_app
from ..logic.users import create_user, UserType
from ..logic.authentication import add_other_authentication


def main(arguments):
    if len(arguments) != 2 or not all(arguments):
        print(__doc__)
        exit(1)
    name, email = arguments
    if '@' not in email[1:-1]:
        print("Error: email must be a valid email address", file=sys.stderr)
        exit(1)
    password = ''.join([('00' + hex(c)[2:])[-2:] for c in os.urandom(16)])
    print("Note: the user will receive the password '{}'".format(password))
    app = create_app()
    with app.app_context():
        user = create_user(name, email, UserType.OTHER)
        add_other_authentication(user.id, name, password)
        print("Success: the user has been created in SampleDB (#{})".format(user.id))



# coding: utf-8
"""

"""

import os
import typing
import pint

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'

ureg = pint.UnitRegistry()
ureg.load_definitions(os.path.join(os.path.dirname(__file__), 'unit_definitions.txt'))


def prettify_units(units: typing.Union[str, ureg.Unit]) -> str:
    """
    Returns a prettified version of the units, if defined, otherwise returns the units unaltered.
    :param units: The pint units or their string representation
    :return: The prettified units
    """
    units = str(units)
    units = {
        'degC': '\xb0C',
        'degree_Celsius': '\xb0C',
        'deg': '\xb0',
        'degree': '\xb0',
        'percent': '%'
    }.get(units, units)
    return units



# coding: utf-8
"""

"""

import typing
from .. import db
from . import errors
from . import actions
from . import groups
from . import users
from . import projects
from . import instruments
from . import settings
from ..models import Permissions, UserActionPermissions, GroupActionPermissions, ProjectActionPermissions, PublicActions, Action

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


def action_is_public(action_id: int) -> bool:
    """
    Return whether an action is public or not.

    :param action_id: the ID of an existing action
    :return: whether the action is public or not
    :raise errors.ActionDoesNotExistError: if no action with the given action
        ID exists
    """
    # ensure that the action can be found
    action = actions.get_action(action_id)
    # common actions are always public
    if action.user_id is None:
        return True
    return PublicActions.query.filter_by(action_id=action_id).first() is not None


def set_action_public(action_id: int, is_public: bool = True) -> None:
    """
    Set that in action is public or not.

    :param action_id: the ID of an existing action
    :param is_public: whether the action should be public or not
    :raise errors.ActionDoesNotExistError: if no action with the given action
        ID exists
    """
    # ensure that the action can be found
    actions.get_action(action_id)
    if not is_public:
        PublicActions.query.filter_by(action_id=action_id).delete()
    elif not action_is_public(action_id):
        db.session.add(PublicActions(action_id=action_id))
    db.session.commit()


def get_action_permissions_for_users(action_id, include_instrument_responsible_users=True, include_groups=True, include_projects=True) -> typing.Dict[int, Permissions]:
    """
    Get permissions for users for a specific action.

    :param action_id: the ID of an existing action
    :param include_instrument_responsible_users: whether instrument responsible user status should be included
    :param include_groups: whether groups that the users are members of should be included
    :param include_projects: whether projects that the users are members of should be included
    :return: a dict mapping users IDs to permissions
    :raise errors.ActionDoesNotExistError: if no action with the given action
        ID exists
    """
    # ensure that the action can be found
    action = actions.get_action(action_id)
    action_permissions = {}
    for user_action_permissions in UserActionPermissions.query.filter_by(action_id=action_id).all():
        action_permissions[user_action_permissions.user_id] = user_action_permissions.permissions
    if include_instrument_responsible_users:
        for user_id in _get_action_responsible_user_ids(action_id):
            action_permissions[user_id] = Permissions.GRANT
    if include_groups:
        for group_action_permissions in GroupActionPermissions.query.filter_by(action_id=action_id).all():
            for user_id in groups.get_group_member_ids(group_action_permissions.group_id):
                if user_id not in action_permissions or action_permissions[user_id] in group_action_permissions.permissions:
                    action_permissions[user_id] = group_action_permissions.permissions
    if include_projects:
        for project_action_permissions in ProjectActionPermissions.query.filter_by(action_id=action_id).all():
            for user_id, permissions in projects.get_project_member_user_ids_and_permissions(project_action_permissions.project_id, include_groups=include_groups).items():
                permissions = min(permissions, project_action_permissions.permissions)
                previous_permissions = action_permissions.get(user_id, Permissions.NONE)
                action_permissions[user_id] = max(previous_permissions, permissions)
    if action.user_id is not None:
        action_permissions[action.user_id] = Permissions.GRANT
    for user_id in action_permissions:
        user = users.get_user(user_id)
        if user.is_readonly:
            action_permissions[user_id] = min(action_permissions[user_id], Permissions.READ)
    return action_permissions


def _get_action_responsible_user_ids(action_id: int) -> typing.List[int]:
    try:
        action = actions.get_action(action_id)
    except errors.ActionDoesNotExistError:
        return []
    if action.instrument_id is None:
        return []
    instrument = instruments.get_instrument(action.instrument_id)
    return [user.id for user in instrument.responsible_users]


def get_action_permissions_for_groups(action_id: int, include_projects=False) -> typing.Dict[int, Permissions]:
    """
    Get permissions for a specific action for groups.

    :param action_id: the ID of an existing action
    :param include_projects: whether projects that the groups are members of should be included
    :return: a dict mapping group IDs to permissions
    :raise errors.ActionDoesNotExistError: if no action with the given action
        ID exists
    """
    # ensure that the action can be found
    actions.get_action(action_id)
    action_permissions = {}
    for group_action_permissions in GroupActionPermissions.query.filter_by(action_id=action_id).all():
        if group_action_permissions.permissions != Permissions.NONE:
            action_permissions[group_action_permissions.group_id] = group_action_permissions.permissions
    if include_projects:
        for project_action_permissions in ProjectActionPermissions.query.filter_by(action_id=action_id).all():
            for group_id, permissions in projects.get_project_member_group_ids_and_permissions(project_action_permissions.project_id).items():
                permissions = min(permissions, project_action_permissions.permissions)
                previous_permissions = action_permissions.get(group_id, Permissions.NONE)
                action_permissions[group_id] = max(previous_permissions, permissions)
    return action_permissions


def get_action_permissions_for_projects(action_id: int) -> typing.Dict[int, Permissions]:
    """
    Get permissions for a specific action for projects.

    :param action_id: the ID of an existing action
    :return: a dict mapping project IDs to permissions
    :raise errors.ActionDoesNotExistError: if no action with the given action
        ID exists
    """
    # ensure that the action can be found
    actions.get_action(action_id)
    action_permissions = {}
    for project_action_permissions in ProjectActionPermissions.query.filter_by(action_id=action_id).all():
        if project_action_permissions.permissions != Permissions.NONE:
            action_permissions[project_action_permissions.project_id] = project_action_permissions.permissions
    return action_permissions


def get_user_action_permissions(action_id, user_id, include_instrument_responsible_users: bool = True, include_groups: bool = True, include_projects: bool = True, include_admin_permissions: bool = True) -> Permissions:
    """
    Get permissions for a specific action for a specific user.

    :param action_id: the ID of an existing action
    :param user_id: the ID of an existing user
    :param include_instrument_responsible_users: whether instrument responsible user status should be included
    :param include_groups: whether groups that the users are members of should be included
    :param include_projects: whether projects that the users are members of should be included
    :param include_admin_permissions: whether admin permissions should be included
    :return: the user's permissions for this action
    :raise errors.ActionDoesNotExistError: if no action with the given action
        ID exists
    :raise errors.UserDoesNotExistError: if no user with the given user ID
        exists
    """
    # ensure that the action can be found
    action = actions.get_action(action_id)
    # ensure that the user can be found
    user = users.get_user(user_id)

    if user.is_readonly:
        max_permissions = Permissions.READ
    else:
        max_permissions = Permissions.GRANT

    # administrators have GRANT permissions if they use admin permissions
    if include_admin_permissions and user.is_admin and settings.get_user_settings(user.id)['USE_ADMIN_PERMISSIONS']:
        return min(Permissions.GRANT, max_permissions)
    # action owners always have GRANT permissions
    if action.user_id == user_id:
        return min(Permissions.GRANT, max_permissions)
    if include_instrument_responsible_users:
        # instrument responsible users always have GRANT permissions for an action
        if user_id in _get_action_responsible_user_ids(action_id):
            return min(Permissions.GRANT, max_permissions)
    # other users might have been granted permissions, either individually or as group or project members
    user_action_permissions = UserActionPermissions.query.filter_by(action_id=action_id, user_id=user_id).first()
    if user_action_permissions is None:
        permissions = Permissions.NONE
    else:
        permissions = user_action_permissions.permissions
    if Permissions.GRANT in permissions:
        return min(permissions, max_permissions)
    if include_groups:
        for group in groups.get_user_groups(user_id):
            group_action_permissions = GroupActionPermissions.query.filter_by(action_id=action_id, group_id=group.id).first()
            if group_action_permissions is not None and permissions in group_action_permissions.permissions:
                permissions = group_action_permissions.permissions
    if Permissions.GRANT in permissions:
        return min(permissions, max_permissions)
    if include_projects:
        for user_project in projects.get_user_projects(user_id, include_groups=include_groups):
            user_project_permissions = projects.get_user_project_permissions(user_project.id, user_id, include_groups=include_groups)
            if user_project_permissions not in permissions:
                project_action_permissions = ProjectActionPermissions.query.filter_by(action_id=action_id, project_id=user_project.id).first()
                if project_action_permissions is not None:
                    permissions = min(user_project_permissions, project_action_permissions.permissions)
    if Permissions.READ in permissions:
        return min(permissions, max_permissions)
    # lastly, the action may be public, so all users have READ permissions
    if action_is_public(action_id):
        return min(Permissions.READ, max_permissions)
    # otherwise the user has no permissions for this action
    return Permissions.NONE


def set_user_action_permissions(action_id: int, user_id: int, permissions: Permissions):
    """
    Set the action permissions for a user.

    Clear the permissions if called with Permissions.NONE

    :param action_id: the ID of an existing action
    :param user_id: the ID of an existing user
    :param permissions: the new permissions
    :raise errors.ActionDoesNotExistError: if no action with the given action
        ID exists
    :raise errors.UserDoesNotExistError: if no user with the given user ID
        exists
    """
    # ensure that the action can be found
    actions.get_action(action_id)
    # ensure that the user can be found
    users.get_user(user_id)
    if permissions == Permissions.NONE:
        UserActionPermissions.query.filter_by(action_id=action_id, user_id=user_id).delete()
    else:
        user_action_permissions = UserActionPermissions.query.filter_by(action_id=action_id, user_id=user_id).first()
        if user_action_permissions is None:
            user_action_permissions = UserActionPermissions(user_id=user_id, action_id=action_id, permissions=permissions)
        else:
            user_action_permissions.permissions = permissions
        db.session.add(user_action_permissions)
    db.session.commit()


def set_group_action_permissions(action_id: int, group_id: int, permissions: Permissions):
    """
    Set the action permissions for a group.

    Clear the permissions if called with Permissions.NONE

    :param action_id: the ID of an existing action
    :param group_id: the ID of an existing group
    :param permissions: the new permissions
    :raise errors.ActionDoesNotExistError: if no action with the given action
        ID exists
    :raise errors.GroupDoesNotExistError: if no group with the given group ID
        exists
    """
    # ensure that the action can be found
    actions.get_action(action_id)
    # ensure that the group can be found
    groups.get_group(group_id)
    if permissions == Permissions.NONE:
        GroupActionPermissions.query.filter_by(action_id=action_id, group_id=group_id).delete()
    else:
        group_action_permissions = GroupActionPermissions.query.filter_by(action_id=action_id, group_id=group_id).first()
        if group_action_permissions is None:
            group_action_permissions = GroupActionPermissions(action_id=action_id, group_id=group_id, permissions=permissions)
        else:
            group_action_permissions.permissions = permissions
        db.session.add(group_action_permissions)
    db.session.commit()


def set_project_action_permissions(action_id: int, project_id: int, permissions: Permissions) -> None:
    """
    Set the action permissions for a group.

    Clear the permissions if called with Permissions.NONE.

    :param action_id: the ID of an existing action
    :param group_id: the ID of an existing group
    :param permissions: the new permissions
    :raise errors.ActionDoesNotExistError: if no action with the given action
        ID exists
    :raise errors.ProjectDoesNotExistError: if no project with the given
        project ID exists
    """
    # ensure that the action can be found
    actions.get_action(action_id)
    # ensure that the project can be found
    projects.get_project(project_id)
    if permissions == Permissions.NONE:
        ProjectActionPermissions.query.filter_by(action_id=action_id, project_id=project_id).delete()
    else:
        project_action_permissions = ProjectActionPermissions.query.filter_by(action_id=action_id, project_id=project_id).first()
        if project_action_permissions is None:
            project_action_permissions = ProjectActionPermissions(project_id=project_id, action_id=action_id, permissions=permissions)
        else:
            project_action_permissions.permissions = permissions
        db.session.add(project_action_permissions)
    db.session.commit()


def get_actions_with_permissions(user_id: int, permissions: Permissions, action_type_id: typing.Optional[int] = None) -> typing.List[Action]:
    """
    Get all actions which a user has the given permissions for.

    Return an empty list if called with Permissions.NONE.

    :param user_id: the ID of an existing user
    :param permissions: the minimum permissions required for the actions for
        the given user
    :param action_type_id: None or the ID of an existing action type
    :return: the actions with the given permissions for the given
        user
    :raise errors.UserDoesNotExistError: if no user with the given user ID
        exists
    :raise errors.ActionTypeDoesNotExistError: when no action type with the
        given action type ID exists
    """
    # ensure that the user can be found
    users.get_user(user_id)
    if permissions == Permissions.NONE:
        return []
    actions_with_permissions = []
    for action in actions.get_actions(action_type_id):
        if permissions in get_user_action_permissions(user_id=user_id, action_id=action.id):
            actions_with_permissions.append(action)
    return actions_with_permissions



# coding: utf-8
"""
Users may upload images to SampleDB for use in Markdown content. As these
files already need to be available to the user while they are editing the
Markdown document, they are will be stored temporarily, only available to that
user. When the Markdown content is then saved, the images that are actually
referenced (via image elements after rendering the Markdown content to HTML)
will be made permanent.

As these images might be referenced from other places afterwards, an image
that was made permanent will not be deleted when its reference is removed from
the Markdown content that it was uploaded for.
"""

import binascii
import os
import typing

from bs4 import BeautifulSoup
import datetime
import flask

from .users import get_user
from ..models.markdown_images import MarkdownImage
from .. import db


def store_temporary_markdown_image(content: bytes, image_file_extension: str, user_id: int) -> str:
    """
    Store a file for use as an image in Markdown content.

    This also triggers the cleanup of previously uploaded temporary files.

    :param content: the binary content of the file
    :param image_file_extension: the file extension that should be used
    :param user_id: the ID of the user who uploaded the file
    :return: the generated file name to identify the image
    :raise errors.UserDoesNotExistError: if no user with the given ID exists
    """
    # ensure the user exists
    get_user(user_id)

    _remove_expired_images()

    # determine random file name
    num_tries = 10
    for _ in range(num_tries):
        file_name = binascii.b2a_hex(os.urandom(32)).decode('ascii') + image_file_extension
        if MarkdownImage.query.filter_by(file_name=file_name).first() is None:
            break
    else:
        # there should be close to no collisions with 32 random bytes, so if
        # several have occurred in a row, something must have been very wrong
        raise RuntimeError()

    image = MarkdownImage(file_name, content, user_id)
    db.session.add(image)
    db.session.commit()
    return file_name


def get_markdown_image(file_name: str, user_id: int) -> typing.Optional[bytes]:
    """
    Return the content of a Markdown image, if it is found.

    If the image is temporary, only the user who uploaded it may access it.

    :param file_name:
    :param user_id:
    :return: the binary file content or None
    """
    image = MarkdownImage.query.filter_by(file_name=file_name).first()
    if image is None:
        return None
    if image.user_id != user_id and not image.permanent:
        return None
    return image.content


def mark_referenced_markdown_images_as_permanent(html_content: str) -> None:
    """
    Find Markdown images referenced in an HTML snippet as permanent.

    :param html_content: the HTML content containing the image references
    """
    for file_name in find_referenced_markdown_images(html_content):
        _mark_markdown_image_permanent(file_name)


def _mark_markdown_image_permanent(file_name: str) -> None:
    image = MarkdownImage.query.filter_by(file_name=file_name, permanent=False).first()
    if image is None:
        return
    image.permanent = True
    db.session.add(image)
    db.session.commit()


def find_referenced_markdown_images(html_content: str) -> typing.Set[str]:
    file_names = set()
    base_image_url = flask.url_for('frontend.markdown_image', file_name='', _external=False)
    soup = BeautifulSoup(html_content, "html.parser")
    for image in soup.findAll('img'):
        image_url = image.get('src')
        if image_url.startswith(base_image_url):
            file_name = image_url[len(base_image_url):]
            file_names.add(file_name)
    return file_names


def _remove_expired_images():
    expiration_datetime = datetime.datetime.utcnow() - datetime.timedelta(days=2)
    expired_images = MarkdownImage.query.filter_by(permanent=False).filter(MarkdownImage.utc_datetime < expiration_datetime).all()
    for expired_image in expired_images:
        db.session.delete(expired_image)
    db.session.commit()



# coding: utf-8
"""

"""

from .. import db
from .groups import Group
from .users import User
from .objects import Objects
from .permissions import Permissions

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


class UserObjectPermissions(db.Model):
    __tablename__ = 'user_object_permissions'

    object_id = db.Column(db.Integer, db.ForeignKey(Objects.object_id_column), nullable=False)
    user_id = db.Column(db.Integer, db.ForeignKey(User.id), nullable=False)
    permissions = db.Column(db.Enum(Permissions), nullable=False, default=Permissions.NONE)

    __table_args__ = (
        db.PrimaryKeyConstraint(object_id, user_id),
        {},
    )


class GroupObjectPermissions(db.Model):
    __tablename__ = 'group_object_permissions'

    object_id = db.Column(db.Integer, db.ForeignKey(Objects.object_id_column), nullable=False)
    group_id = db.Column(db.Integer, db.ForeignKey(Group.id, ondelete="CASCADE"), nullable=False)
    permissions = db.Column(db.Enum(Permissions), nullable=False, default=Permissions.NONE)

    __table_args__ = (
        db.PrimaryKeyConstraint(object_id, group_id),
        {},
    )


class ProjectObjectPermissions(db.Model):
    __tablename__ = 'project_object_permissions'

    object_id = db.Column(db.Integer, db.ForeignKey(Objects.object_id_column), nullable=False)
    project_id = db.Column(db.Integer, db.ForeignKey('projects.id', ondelete="CASCADE"), nullable=False)
    permissions = db.Column(db.Enum(Permissions), nullable=False, default=Permissions.NONE)

    __table_args__ = (
        db.PrimaryKeyConstraint(object_id, project_id),
        {},
    )


class PublicObjects(db.Model):
    __tablename__ = 'public_objects'

    object_id = db.Column(db.Integer, db.ForeignKey(Objects.object_id_column), primary_key=True)


class DefaultUserPermissions(db.Model):
    __tablename__ = 'default_user_permissions'

    creator_id = db.Column(db.Integer, db.ForeignKey(User.id), nullable=False)
    user_id = db.Column(db.Integer, db.ForeignKey(User.id), nullable=False)
    permissions = db.Column(db.Enum(Permissions), nullable=False, default=Permissions.NONE)

    __table_args__ = (
        db.PrimaryKeyConstraint(creator_id, user_id),
        {},
    )


class DefaultGroupPermissions(db.Model):
    __tablename__ = 'default_group_permissions'

    creator_id = db.Column(db.Integer, db.ForeignKey(User.id), nullable=False)
    group_id = db.Column(db.Integer, db.ForeignKey(Group.id, ondelete="CASCADE"), nullable=False)
    permissions = db.Column(db.Enum(Permissions), nullable=False, default=Permissions.NONE)

    __table_args__ = (
        db.PrimaryKeyConstraint(creator_id, group_id),
        {},
    )


class DefaultProjectPermissions(db.Model):
    __tablename__ = 'default_project_permissions'

    creator_id = db.Column(db.Integer, db.ForeignKey(User.id), nullable=False)
    project_id = db.Column(db.Integer, db.ForeignKey('projects.id', ondelete="CASCADE"), nullable=False)
    permissions = db.Column(db.Enum(Permissions), nullable=False, default=Permissions.NONE)

    __table_args__ = (
        db.PrimaryKeyConstraint(creator_id, project_id),
        {},
    )


class DefaultPublicPermissions(db.Model):
    __tablename__ = 'default_public_permissions'

    creator_id = db.Column(db.Integer, db.ForeignKey(User.id), primary_key=True)
    is_public = db.Column(db.Boolean, default=False, nullable=False)



import bcrypt
import typing
import flask

from .. import logic, db
from .ldap import validate_user, create_user_from_ldap, is_ldap_configured
from ..models import Authentication, AuthenticationType, User
from . import errors, api_log


def _hash_password(password: str) -> str:
    return bcrypt.hashpw(password.encode('utf-8'), bcrypt.gensalt()).decode('utf-8')


def _validate_password_hash(password: str, password_hash: str) -> bool:
    return bcrypt.checkpw(password.encode('utf-8'), password_hash.encode('utf-8'))


def _validate_password_authentication(authentication_method: Authentication, password: str) -> bool:
    password_hash = authentication_method.login['bcrypt_hash']
    return _validate_password_hash(password, password_hash)


def _add_password_authentication(user_id: int, login: str, password: str, authentication_type: AuthenticationType, confirmed: bool = True) -> None:
    login = login.lower().strip()
    if Authentication.query.filter(Authentication.login['login'].astext == login).first():
        raise errors.AuthenticationMethodAlreadyExists('An authentication method with this login already exists')
    authentication = Authentication(
        login={'login': login, 'bcrypt_hash': _hash_password(password)},
        authentication_type=authentication_type,
        confirmed=confirmed,
        user_id=user_id
    )
    db.session.add(authentication)
    db.session.commit()


def add_other_authentication(user_id: int, name: str, password: str, confirmed: bool = True) -> None:
    """
    Add an authentication method of type OTHER for a given user.

    :param user_id: the ID of an existing user
    :param name: the name to use during authentication
    :param password: the password
    :param confirmed: whether the authentication method has been confirmed already
    """
    _add_password_authentication(user_id, name, password, AuthenticationType.OTHER, confirmed)


def add_email_authentication(user_id: int, email: str, password: str, confirmed: bool = True) -> None:
    """
    Add an authentication method of type EMAIL for a given user.

    :param user_id: the ID of an existing user
    :param email: the email to use during authentication
    :param password: the password
    :param confirmed: whether the authentication method has been confirmed already
    """
    email = email.lower().strip()
    if '@' not in email[1:-1]:
        raise errors.AuthenticationMethodWrong('Login must be a valid email address')
    _add_password_authentication(user_id, email, password, AuthenticationType.EMAIL, confirmed)
    if not confirmed:
        logic.utils.send_email_confirmation_email(
            email=email,
            user_id=user_id,
            salt='add_login'
        )


def add_ldap_authentication(user_id: int, ldap_uid: str, password: str, confirmed: bool = True) -> None:
    """
    Add an authentication method of type LDAP for a given user.

    :param user_id: the ID of an existing user
    :param ldap_uid: the LDAP uid to use during authentication
    :param confirmed: whether the authentication method has been confirmed already
    """
    ldap_uid = ldap_uid.lower().strip()
    if Authentication.query.filter(Authentication.login['login'].astext == ldap_uid).first():
        raise errors.AuthenticationMethodAlreadyExists('An authentication method with this login already exists')
    if Authentication.query.filter(Authentication.type == AuthenticationType.LDAP, Authentication.user_id == user_id).first():
        raise errors.LDAPAccountAlreadyExistError('An LDAP-based authentication method already exists for this user')
    if not validate_user(ldap_uid, password):
        raise errors.LDAPAccountOrPasswordWrongError('Ldap login or password wrong')
    authentication = Authentication(
        login={'login': ldap_uid},
        authentication_type=AuthenticationType.LDAP,
        confirmed=confirmed,
        user_id=user_id
    )
    db.session.add(authentication)
    db.session.commit()


def add_api_token(user_id: int, api_token: str, description: str) -> None:
    """
    Add an API token as an authentication method for a given user.

    :param user_id: the ID of an existing user
    :param api_token: the api token as a 64 character string
    :param description: a description for the token
    """

    assert len(api_token) == 64
    api_token = api_token.lower()
    # split into a short part (8 hex digits / 4 bytes) for identification and a long part for authentication
    login, password = api_token[:8], api_token[8:]
    authentication = Authentication(
        login={
            'login': login,
            'bcrypt_hash': _hash_password(password),
            'description': description
        },
        authentication_type=AuthenticationType.API_TOKEN,
        confirmed=True,
        user_id=user_id
    )
    db.session.add(authentication)
    db.session.commit()


def login(login: str, password: str) -> typing.Optional[User]:
    """
    Authenticate a user and create an LDAP based user if necessary.

    :param login: the name, email or LDAP uid to use during authentication
    :param password: the password
    :return: the user or None
    """
    # convert to lower case to enforce case insensitivity
    login = login.lower().strip()
    # filter email + password or username + password or username (ldap)
    authentication_methods = Authentication.query.filter(
        db.or_(
            db.and_(Authentication.login['login'].astext == login,
                    Authentication.type == AuthenticationType.EMAIL),
            db.and_(Authentication.login['login'].astext == login,
                    Authentication.type == AuthenticationType.LDAP),
            db.and_(Authentication.login['login'].astext == login,
                    Authentication.type == AuthenticationType.OTHER)
        )
    ).all()

    for authentication_method in authentication_methods:
        if not authentication_method.confirmed:
            continue
        if authentication_method.type == AuthenticationType.LDAP and is_ldap_configured():
            if validate_user(login, password):
                return authentication_method.user
        elif authentication_method.type in {AuthenticationType.EMAIL, AuthenticationType.OTHER}:
            if _validate_password_authentication(authentication_method, password):
                return authentication_method.user

    # no matching authentication method in db
    if not authentication_methods and '@' not in login and is_ldap_configured():
        # try to authenticate against ldap, if login is no email
        if not validate_user(login, password):
            return None

        user = create_user_from_ldap(login)
        if user is None:
            return None
        try:
            add_ldap_authentication(user.id, login, password)
        except errors.AuthenticationMethodAlreadyExists:
            # user might have been created in the background already due to concurrent, duplicate requests
            pass
        return user
    return None


def login_via_api_token(api_token: str) -> typing.Optional[User]:
    """
    Authenticate a user using an API token.

    :param api_token: the API token to use during authentication
    :return: the user or None
    """
    # convert to lower case to enforce case insensitivity
    api_token = api_token.lower().strip()
    username, password = api_token[:8], api_token[8:]
    authentication_methods = Authentication.query.filter(
        db.and_(Authentication.login['login'].astext == username,
                Authentication.type == AuthenticationType.API_TOKEN)
    ).all()

    for authentication_method in authentication_methods:
        if not authentication_method.confirmed:
            continue
        if _validate_password_authentication(authentication_method, password):
            api_log.create_log_entry(authentication_method.id, getattr(api_log.HTTPMethod, flask.request.method, api_log.HTTPMethod.OTHER), flask.request.path)
            return authentication_method.user
    return None


def add_authentication_method(user_id: int, login: str, password: str, authentication_type: AuthenticationType) -> bool:
    """
    Add an authentication method for a user.

    :param user_id: the ID of an existing user
    :param login: the name, email or LDAP uid to use
    :param password: the password
    :param authentication_type: the type of authentication
    :return: whether or not the authentication method was added
    :raise errors.AuthenticationMethodAlreadyExists: if the authentication method exists already
    :raise errors.AuthenticationMethodWrong: if the authentication method was used in a wrong way
    :raise errors.LdapAccountAlreadyExist: if a user with this LDAP account already exists
    :raise errors.LdapAccountOrPasswordWrong: if the LDAP account and password could not be validated
    """
    if password is None or password == '':
        return False
    if login is None or login == '':
        return False
    # convert to lower case to enforce case insensitivity
    login = login.lower().strip()

    if authentication_type == AuthenticationType.LDAP:
        add_ldap_authentication(user_id, login, password)
        return True
    if authentication_type == AuthenticationType.EMAIL:
        add_email_authentication(user_id, login, password, False)
        return True
    return False


def remove_authentication_method(authentication_method_id: int) -> bool:
    """
    Remove an authentication method.

    :param authentication_method_id: the ID of an existing authentication method
    :return:
    """
    authentication_method = Authentication.query.filter(Authentication.id == authentication_method_id).first()
    if authentication_method is None:
        return False
    if authentication_method.type != AuthenticationType.API_TOKEN:
        authentication_methods_count = Authentication.query.filter(Authentication.user_id == authentication_method.user_id, Authentication.type != AuthenticationType.API_TOKEN).count()
        if authentication_methods_count <= 1:
            raise errors.OnlyOneAuthenticationMethod('one authentication-method must at least exist, delete not possible')
    db.session.delete(authentication_method)
    db.session.commit()
    return True


def change_password_in_authentication_method(authentication_method_id: int, password: str) -> bool:
    """
    Change the password for a password-based authentication method.

    :param authentication_method_id: the ID of an authentication method
    :param password: the new password
    :return: whether the password was changed
    """
    if password is None or password == '':
        return False
    if authentication_method_id is None or authentication_method_id <= 0:
        return False
    authentication_method = Authentication.query.filter(Authentication.id == authentication_method_id).first()
    if authentication_method is None:
        return False
    if authentication_method.type not in {AuthenticationType.EMAIL, AuthenticationType.OTHER}:
        return False
    authentication_method.login = {'login': authentication_method.login['login'], 'bcrypt_hash': _hash_password(password)}
    db.session.add(authentication_method)
    db.session.commit()
    return True


def is_login_available(login: str) -> bool:
    """
    Return whether or not a login is still available for use.

    For password-based authentication, the login can only be used once.

    :param login: the login for which to check availability
    :return: whether the login is available
    """
    return Authentication.query.filter(Authentication.login['login'].astext == login).first() is None



# coding: utf-8
"""

"""

import datetime
import typing

import flask
import flask_login
from flask_wtf import FlaskForm

from wtforms.fields import IntegerField
from wtforms.validators import InputRequired

from .. import frontend
from ...logic import errors, users, groups, projects, object_permissions, locations, instrument_log_entries, instruments
from ...logic.notifications import get_notification, get_notifications, mark_notification_as_read, delete_notification, NotificationType


class DeleteAllNotificationsForm(FlaskForm):
    delete_all_notifications_up_to_id = IntegerField(validators=[InputRequired()])


class MarkAllNotificationsAsReadForm(FlaskForm):
    mark_all_notifications_as_read_up_to_id = IntegerField(validators=[InputRequired()])


class DeleteNotificationForm(FlaskForm):
    delete_notification = IntegerField(validators=[InputRequired()])


class MarkNotificationAsReadForm(FlaskForm):
    mark_notification_read = IntegerField(validators=[InputRequired()])


@frontend.route('/users/me/notifications')
@flask_login.login_required
def current_user_notifications():
    return flask.redirect(flask.url_for('.notifications', user_id=flask_login.current_user.id))


@frontend.route('/users/<int:user_id>/notifications', methods=['GET', 'POST'])
@flask_login.login_required
def notifications(user_id):
    if user_id != flask_login.current_user.id:
        return flask.abort(403)

    delete_all_notifications_form = DeleteAllNotificationsForm()
    mark_all_notifications_as_read_form = MarkAllNotificationsAsReadForm()
    delete_notification_form = DeleteNotificationForm()
    mark_notification_as_read_form = MarkNotificationAsReadForm()

    if delete_all_notifications_form.validate_on_submit():
        max_known_notification_id = int(delete_all_notifications_form.delete_all_notifications_up_to_id.data)
        notifications = get_notifications(user_id)
        for notification in notifications:
            if notification.id <= max_known_notification_id:
                try:
                    delete_notification(notification.id)
                except errors.NotificationDoesNotExistError:
                    continue
        flask.flash('The notifications have been deleted.', 'success')
        return flask.redirect(flask.url_for('.notifications', user_id=user_id))

    if mark_all_notifications_as_read_form.validate_on_submit():
        max_known_notification_id = int(mark_all_notifications_as_read_form.mark_all_notifications_as_read_up_to_id.data)
        notifications = get_notifications(user_id, unread_only=True)
        for notification in notifications:
            if notification.id <= max_known_notification_id:
                try:
                    mark_notification_as_read(notification.id)
                except errors.NotificationDoesNotExistError:
                    continue
        flask.flash('The notifications have been marked as read.', 'success')
        return flask.redirect(flask.url_for('.notifications', user_id=user_id))

    if delete_notification_form.validate_on_submit():
        notification_id = int(delete_notification_form.delete_notification.data)
        try:
            notification = get_notification(notification_id)
            if notification.user_id == flask_login.current_user.id:
                delete_notification(notification_id)
        except errors.NotificationDoesNotExistError:
            flask.flash('This notification does not exist.', 'error')
        else:
            flask.flash('The notification has been deleted.', 'success')
        return flask.redirect(flask.url_for('.notifications', user_id=user_id))

    if mark_notification_as_read_form.validate_on_submit():
        notification_id = int(mark_notification_as_read_form.mark_notification_read.data)
        try:
            notification = get_notification(notification_id)
            if notification.user_id == flask_login.current_user.id:
                mark_notification_as_read(notification_id)
        except errors.NotificationDoesNotExistError:
            flask.flash('This notification does not exist.', 'error')
        else:
            flask.flash('The notification has been marked as read.', 'success')
        return flask.redirect(flask.url_for('.notifications', user_id=user_id))

    notifications = get_notifications(user_id)
    if notifications:
        max_known_notification_id = max(notification.id for notification in notifications)
    else:
        max_known_notification_id = 0
    return flask.render_template(
        'notifications.html',
        NotificationType=NotificationType,
        notifications=notifications,
        max_known_notification_id=max_known_notification_id,
        delete_all_notifications_form=delete_all_notifications_form,
        mark_all_notifications_as_read_form=mark_all_notifications_as_read_form,
        delete_notification_form=delete_notification_form,
        mark_notification_as_read_form=mark_notification_as_read_form,
        get_user=users.get_user,
        get_group=_safe_get_group,
        is_group_member=_is_group_member,
        get_project=_safe_get_project,
        get_instrument=_safe_get_instrument,
        get_instrument_log_entry=_safe_get_instrument_log_entry,
        is_project_member=_is_project_member,
        get_user_object_permissions=object_permissions.get_user_object_permissions,
        Permissions=object_permissions.Permissions,
        object_location_assignment_is_confirmed=lambda object_location_assignment_id: locations.get_object_location_assignment(object_location_assignment_id).confirmed,
        datetime=datetime
    )


def _safe_get_group(group_id: int) -> typing.Optional[groups.Group]:
    try:
        return groups.get_group(group_id)
    except errors.GroupDoesNotExistError:
        return None


def _is_group_member(user_id: int, group_id: int) -> bool:
    user_groups = groups.get_user_groups(user_id)
    if not user_groups:
        return False
    return any(group_id == group.id for group in user_groups)


def _safe_get_project(project_id: int) -> typing.Optional[projects.Project]:
    try:
        return projects.get_project(project_id)
    except errors.ProjectDoesNotExistError:
        return None


def _is_project_member(user_id: int, project_id: int) -> bool:
    user_projects = projects.get_user_projects(user_id)
    if not user_projects:
        return False
    return any(project_id == project.id for project in user_projects)


def _safe_get_instrument(instrument_id: int) -> typing.Optional[instruments.Instrument]:
    try:
        return instruments.get_instrument(instrument_id)
    except errors.InstrumentDoesNotExistError:
        return None


def _safe_get_instrument_log_entry(instrument_log_entry_id: int) -> typing.Optional[instrument_log_entries.InstrumentLogEntry]:
    try:
        return instrument_log_entries.get_instrument_log_entry(instrument_log_entry_id)
    except errors.InstrumentLogEntryDoesNotExistError:
        return None



# coding: utf-8
"""
Logic module for management of instruments

Instruments are used to group actions together and provide instrument
responsible users with permissions for samples or measurements produced
using the instrument.

Similar to actions, instruments cannot be deleted. However, all information
about an instrument may be altered.
"""

import typing

from .. import db
from ..models import Instrument
from ..models.instruments import instrument_user_association_table
from . import users, errors


def create_instrument(
        name: str,
        description: str,
        description_as_html: typing.Optional[str] = None,
        users_can_create_log_entries: bool = False,
        users_can_view_log_entries: bool = False,
        notes: str = '',
        notes_as_html: typing.Optional[str] = None,
        create_log_entry_default: bool = False,
        is_hidden: bool = False
) -> Instrument:
    """
    Creates a new instrument with the given name and description.

    :param name: the name of the instrument
    :param description: a (possibly empty) description of the instrument
    :param description_as_html: None or the description as HTML
    :param users_can_create_log_entries: whether or not users can create log
        entries for this instrument
    :param users_can_view_log_entries: whether or not users can view the log
        entries for this instrument
    :param notes: the notes shown to instrument responsible users
    :param notes_as_html: None or the notes as HTML
    :param create_log_entry_default: the default for whether or not a log
        entry should be created during object creation by instrument scientists
    :param is_hidden: whether or not this instrument is hidden
    :return: the new instrument
    """
    instrument = Instrument(
        name=name,
        description=description,
        description_as_html=description_as_html,
        users_can_create_log_entries=users_can_create_log_entries,
        users_can_view_log_entries=users_can_view_log_entries,
        notes=notes,
        notes_as_html=notes_as_html,
        create_log_entry_default=create_log_entry_default,
        is_hidden=is_hidden
    )
    db.session.add(instrument)
    db.session.commit()
    return instrument


def get_instruments() -> typing.List[Instrument]:
    """
    Returns all instruments.

    :return: the list of instruments
    """
    return Instrument.query.all()


def get_instrument(instrument_id: int) -> Instrument:
    """
    Returns the instrument with the given instrument ID.

    :param instrument_id: the ID of an existing instrument
    :return: the instrument
    :raise errors.InstrumentDoesNotExistError: when no instrument with the
        given instrument ID exists
    """
    instrument = Instrument.query.get(instrument_id)
    if instrument is None:
        raise errors.InstrumentDoesNotExistError()
    return instrument


def update_instrument(
        instrument_id: int,
        name: str,
        description: str,
        description_as_html: typing.Optional[str] = None,
        users_can_create_log_entries: bool = False,
        users_can_view_log_entries: bool = False,
        notes: str = '',
        notes_as_html: typing.Optional[str] = None,
        create_log_entry_default: bool = False,
        is_hidden: typing.Optional[bool] = None
) -> None:
    """
    Updates the instrument name and description.

    :param instrument_id: the ID of an existing instrument
    :param name: the new name of the instrument
    :param description: the new (possibly empty) description of the instrument
    :param description_as_html: None or the description as HTML
    :param users_can_create_log_entries: whether or not users can create log
        entries for this instrument
    :param users_can_view_log_entries: whether or not users can view the log
        entries for this instrument
    :param notes: the notes shown to instrument responsible users
    :param notes_as_html: None or the notes as HTML
    :param create_log_entry_default: the default for whether or not a log
        entry should be created during object creation by instrument scientists
    :param is_hidden: None or whether or not this instrument is hidden
    :raise errors.InstrumentDoesNotExistError: when no instrument with the
        given instrument ID exists
    """
    instrument = Instrument.query.get(instrument_id)
    if instrument is None:
        raise errors.InstrumentDoesNotExistError()
    instrument.name = name
    instrument.description = description
    instrument.description_as_html = description_as_html
    instrument.users_can_create_log_entries = users_can_create_log_entries
    instrument.users_can_view_log_entries = users_can_view_log_entries
    instrument.notes = notes
    instrument.notes_as_html = notes_as_html
    instrument.create_log_entry_default = create_log_entry_default
    if is_hidden is not None:
        instrument.is_hidden = is_hidden
    db.session.add(instrument)
    db.session.commit()


def add_instrument_responsible_user(instrument_id: int, user_id: int) -> None:
    """
    Adds the user with the given user ID to the list of instrument responsible
    users.

    :param instrument_id: the ID of an existing instrument
    :param user_id: the ID of an existing user
    :raise errors.InstrumentDoesNotExistError: when no instrument with the
        given instrument ID exists
    :raise errors.UserDoesNotExistError: when no user with the given user ID
        exists
    :raise errors.UserAlreadyResponsibleForInstrumentError: when the user is
        already responsible for the instrument
    """
    instrument = Instrument.query.get(instrument_id)
    if instrument is None:
        raise errors.InstrumentDoesNotExistError()
    user = users.get_user(user_id)
    if user in instrument.responsible_users:
        raise errors.UserAlreadyResponsibleForInstrumentError()
    instrument.responsible_users.append(user)
    db.session.add(instrument)
    db.session.commit()


def remove_instrument_responsible_user(instrument_id: int, user_id: int) -> None:
    """
    Removes the user with the given user ID from the list of instrument
    responsible users.

    :param instrument_id: the ID of an existing instrument
    :param user_id: the ID of an existing user
    :raise errors.InstrumentDoesNotExistError: when no instrument with the
        given instrument ID exists
    :raise errors.UserDoesNotExistError: when no user with the given user ID
        exists
    :raise errors.UserNotResponsibleForInstrumentError: when the user is not
        responsible for the instrument
    """
    instrument = Instrument.query.get(instrument_id)
    if instrument is None:
        raise errors.InstrumentDoesNotExistError()
    user = users.get_user(user_id)
    if user not in instrument.responsible_users:
        raise errors.UserNotResponsibleForInstrumentError()
    instrument.responsible_users.remove(user)
    db.session.add(instrument)
    db.session.commit()


def set_instrument_responsible_users(instrument_id: int, user_ids: typing.List[int]) -> None:
    """
    Set the list of instrument responsible users for a an instrument.

    :param instrument_id: the ID of an existing instrument
    :param user_ids: the IDs of existing users
    :raise errors.InstrumentDoesNotExistError: when no instrument with the
        given instrument ID exists
    :raise errors.UserDoesNotExistError: when no user with one of the given
        user IDs exists
    """
    instrument = Instrument.query.get(instrument_id)
    if instrument is None:
        raise errors.InstrumentDoesNotExistError()
    instrument.responsible_users.clear()
    for user_id in user_ids:
        user = users.get_user(user_id)
        instrument.responsible_users.append(user)
    db.session.add(instrument)
    db.session.commit()


def get_user_instruments(user_id: int, exclude_hidden: bool = False) -> typing.List[int]:
    """
    Get a list of instruments a user with a given user ID is responsible for.

    :param user_id: the ID of an existing user
    :param exclude_hidden: whether hidden instruments should be excluded
    :return: a list of instrument IDs
    :raise errors.UserDoesNotExistError: when no user with the given user ID
        exists
    """
    # ensure that the user exists
    users.get_user(user_id)
    instrument_id_query = db.session.query(instrument_user_association_table.c.instrument_id).filter(instrument_user_association_table.c.user_id == user_id)
    if exclude_hidden:
        instrument_id_query = instrument_id_query.join(
            Instrument,
            instrument_user_association_table.c.instrument_id == Instrument.id
        ).filter(Instrument.is_hidden == db.false())
    instrument_ids = [
        instrument_user_association[0]
        for instrument_user_association in instrument_id_query.order_by(instrument_user_association_table.c.instrument_id).all()
    ]
    return instrument_ids



# coding: utf-8
"""

"""

from http import HTTPStatus
import flask
import flask_login
from flask_wtf import FlaskForm
from wtforms import StringField, PasswordField
from wtforms.validators import Email, Length, InputRequired, ValidationError

from .. import frontend
from ...logic import users, authentication


class CreateOtherUserForm(FlaskForm):
    name = StringField(validators=[InputRequired()])
    email = StringField(validators=[Email()])
    password = PasswordField(validators=[Length(min=3)])

    def validate_name(form, field):
        if len(field.data) < 1:
            raise ValidationError('Please enter a valid user name.')
        if not all(c in 'abcdefghijklmnopqrstuvwxyz0123456789_' for c in field.data):
            raise ValidationError('The name may only contain lower case characters, digits and underscores.')
        if not authentication.is_login_available(field.data):
            raise ValidationError('This name is already being used.')


@frontend.route('/users/create_other_user', methods=['GET', 'POST'])
@flask_login.login_required
def create_other_user():
    if not flask_login.current_user.is_admin:
        return flask.abort(HTTPStatus.UNAUTHORIZED)
    create_other_user_form = CreateOtherUserForm()
    if create_other_user_form.validate_on_submit():
        user = users.create_user(
            name=create_other_user_form.name.data,
            email=create_other_user_form.email.data,
            type=users.UserType.OTHER
        )
        authentication.add_other_authentication(
            user_id=user.id,
            name=create_other_user_form.name.data,
            password=create_other_user_form.password.data
        )
        flask.flash('The user has been created successfully.', 'success')
        return flask.redirect(flask.url_for('.user_profile', user_id=user.id))
    return flask.render_template(
        'create_other_user.html',
        create_other_user_form=create_other_user_form
    )



# coding: utf-8
"""
Authentication functions for the SampleDB RESTful API.
"""

import flask

from flask_httpauth import HTTPBasicAuth, HTTPTokenAuth, MultiAuth

from ...logic.authentication import login, login_via_api_token
from ...logic.object_permissions import Permissions
from ...utils import object_permissions_required as object_permissions_required_generic

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'

http_basic_auth = HTTPBasicAuth()
http_token_auth = HTTPTokenAuth(scheme='Bearer')
multi_auth = MultiAuth(http_basic_auth, http_token_auth)


@http_token_auth.verify_token
def verify_token(api_token):
    if not api_token:
        return None
    user = login_via_api_token(api_token)
    if user is None or not user.is_active:
        return None
    flask.g.user = user
    return user


@http_basic_auth.verify_password
def verify_password(username, password):
    if not username:
        return None
    user = login(username, password)
    if user is None or not user.is_active:
        return None
    flask.g.user = user
    return user


def object_permissions_required(permissions: Permissions):
    """
    Only allow access to a route it the user has the required permissions.

    Wrapper around the more generic sampledb.utils.object_permissions_required
    for use with the http_basic_auth object from this module.

    :param permissions: the required object permissions
    """
    return object_permissions_required_generic(permissions, multi_auth, lambda: flask.g.user.id)



# coding: utf-8
"""

"""

import flask

from . import frontend

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


@frontend.route('/')
def index():
    return flask.render_template('index.html')



# coding: utf-8
"""

"""

import datetime
import typing
from .errors import ObjectDoesNotExistError
from .users import get_user
from .object_permissions import get_user_object_permissions, Permissions
from ..models import UserLogEntry, UserLogEntryType
from .. import db

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


def get_user_log_entries(user_id: int, as_user_id: typing.Optional[int] = None) -> typing.List[UserLogEntry]:
    user_log_entries = UserLogEntry.query.filter_by(user_id=user_id).order_by(db.desc(UserLogEntry.utc_datetime)).all()
    if as_user_id is None or as_user_id == user_id or get_user(as_user_id).is_admin:
        return user_log_entries
    visible_user_log_entries = []
    for user_log_entry in user_log_entries:
        if 'object_id' in user_log_entry.data:
            object_id = user_log_entry.data['object_id']
            try:
                if Permissions.READ in get_user_object_permissions(user_id=as_user_id, object_id=object_id):
                    visible_user_log_entries.append(user_log_entry)
            except ObjectDoesNotExistError:
                pass
        elif 'object_ids' in user_log_entry.data:
            object_ids = user_log_entry.data['object_ids']
            for object_id in object_ids:
                try:
                    if Permissions.READ in get_user_object_permissions(user_id=as_user_id, object_id=object_id):
                        visible_user_log_entries.append(user_log_entry)
                        break
                except ObjectDoesNotExistError:
                    pass
    return visible_user_log_entries


def get_user_related_object_ids(user_id: int) -> typing.Set[int]:
    """
    Return a set of IDs of all objects related to a given user.

    :param user_id: the ID of an existing user
    :return: a set of object IDs related to the user
    """
    user_log_entries = UserLogEntry.query.filter_by(user_id=user_id).all()
    user_related_object_ids = set()
    for user_log_entry in user_log_entries:
        if 'object_id' in user_log_entry.data:
            user_related_object_ids.add(user_log_entry.data['object_id'])
        elif 'object_ids' in user_log_entry.data:
            user_related_object_ids.update(set(user_log_entry.data['object_ids']))
    return user_related_object_ids


def _store_new_log_entry(type: UserLogEntryType, user_id: int, data: dict):
    user_log_entry = UserLogEntry(
        type=type,
        user_id=user_id,
        data=data,
        utc_datetime=datetime.datetime.utcnow()
    )
    db.session.add(user_log_entry)
    db.session.commit()


def create_object(user_id: int, object_id: int):
    _store_new_log_entry(
        type=UserLogEntryType.CREATE_OBJECT,
        user_id=user_id,
        data={
            'object_id': object_id
        }
    )


def edit_object(user_id: int, object_id: int, version_id: int):
    _store_new_log_entry(
        type=UserLogEntryType.EDIT_OBJECT,
        user_id=user_id,
        data={
            'object_id': object_id,
            'version_id': version_id
        }
    )


def restore_object_version(user_id: int, object_id: int, version_id: int, restored_version_id: int):
    _store_new_log_entry(
        type=UserLogEntryType.RESTORE_OBJECT_VERSION,
        user_id=user_id,
        data={
            'object_id': object_id,
            'version_id': version_id,
            'restored_version_id': restored_version_id
        }
    )


def edit_object_permissions(user_id: int, object_id: int):
    _store_new_log_entry(
        type=UserLogEntryType.EDIT_OBJECT_PERMISSIONS,
        user_id=user_id,
        data={
            'object_id': object_id
        }
    )


def register_user(user_id: int, email: str):
    _store_new_log_entry(
        type=UserLogEntryType.REGISTER_USER,
        user_id=user_id,
        data={
            'email': email
        }
    )


def invite_user(user_id: int, email: str):
    _store_new_log_entry(
        type=UserLogEntryType.INVITE_USER,
        user_id=user_id,
        data={
            'email': email
        }
    )


def edit_user_preferences(user_id: int):
    _store_new_log_entry(
        type=UserLogEntryType.EDIT_USER_PREFERENCES,
        user_id=user_id,
        data={}
    )


def post_comment(user_id: int, object_id: int, comment_id: int):
    _store_new_log_entry(
        type=UserLogEntryType.POST_COMMENT,
        user_id=user_id,
        data={
            'object_id': object_id,
            'comment_id': comment_id
        }
    )


def upload_file(user_id: int, object_id: int, file_id: int):
    _store_new_log_entry(
        type=UserLogEntryType.UPLOAD_FILE,
        user_id=user_id,
        data={
            'object_id': object_id,
            'file_id': file_id
        }
    )


def create_batch(user_id: int, batch_object_ids: typing.List[int]):
    _store_new_log_entry(
        type=UserLogEntryType.CREATE_BATCH,
        user_id=user_id,
        data={
            'object_ids': batch_object_ids
        }
    )


def assign_location(user_id: int, object_location_assignment_id: int):
    _store_new_log_entry(
        type=UserLogEntryType.ASSIGN_LOCATION,
        user_id=user_id,
        data={
            'object_location_assignment_id': object_location_assignment_id
        }
    )


def create_location(user_id: int, location_id: int):
    _store_new_log_entry(
        type=UserLogEntryType.CREATE_LOCATION,
        user_id=user_id,
        data={
            'location_id': location_id
        }
    )


def update_location(user_id: int, location_id: int):
    _store_new_log_entry(
        type=UserLogEntryType.UPDATE_LOCATION,
        user_id=user_id,
        data={
            'location_id': location_id
        }
    )


def link_publication(
        user_id: int,
        object_id: int,
        doi: str,
        title: typing.Optional[str] = None,
        object_name: typing.Optional[str] = None
):
    _store_new_log_entry(
        type=UserLogEntryType.LINK_PUBLICATION,
        user_id=user_id,
        data={
            'object_id': object_id,
            'doi': doi,
            'title': title,
            'object_name': object_name
        }
    )


def create_instrument_log_entry(user_id: int, instrument_id: int, instrument_log_entry_id: int):
    _store_new_log_entry(
        type=UserLogEntryType.CREATE_INSTRUMENT_LOG_ENTRY,
        user_id=user_id,
        data={
            'instrument_id': instrument_id,
            'instrument_log_entry_id': instrument_log_entry_id
        }
    )



# coding: utf-8
"""
RESTful API for SampleDB
"""

from flask_restful import Resource

from .authentication import multi_auth
from ...logic.actions import get_action_types, get_action_type
from ...logic import errors

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


def action_type_to_json(action_type):
    return {
        'type_id': action_type.id,
        'name': action_type.name,
        'object_name': action_type.object_name,
        'admin_only': action_type.admin_only
    }


class ActionType(Resource):
    @multi_auth.login_required
    def get(self, type_id: int):
        try:
            action_type = get_action_type(action_type_id=type_id)
        except errors.ActionTypeDoesNotExistError:
            return {
                "message": "action type {} does not exist".format(type_id)
            }, 404
        return action_type_to_json(action_type)


class ActionTypes(Resource):
    @multi_auth.login_required
    def get(self):
        action_types = get_action_types()
        return [action_type_to_json(action_type) for action_type in action_types]



# coding: utf-8
"""

"""

import importlib
import os
import typing


def should_skip_by_index(db: typing.Any, index: int) -> bool:
    """
    Returns whether or not a migration should be skipped due to its index.

    :param db: the database
    :param index: the migration index to check
    :return: whether or not the migration should be skipped
    """
    # migration 0 creates the migration_index table, so it cannot be skipped
    # by index as the table might not exist yet
    if index == 0:
        return False

    return db.session.execute(
        """
        SELECT migration_index
        FROM migration_index
        WHERE migration_index >= :index
        """,
        {'index': index}
    ).first() is not None


def update_migration_index(db: typing.Any, index: int) -> None:
    """
    Set the database's migration index to the given index.

    This function also ensures that the migration index can only ever be
    increased and fails with an assertion error otherwise.

    :param db: the database
    :param index: the new migration index
    """
    # migration 0 creates the migration_index table and sets it to 0
    if index == 0:
        return

    assert db.session.execute(
        """
        UPDATE migration_index
        SET migration_index = :index
        WHERE migration_index < :index
        """,
        {"index": index}
    ).rowcount == 1


def find_migrations() -> typing.List[typing.Tuple[int, str, typing.Callable[[typing.Any], bool]]]:
    """
    Finds migrations and returns them sorted by their index.

    :return: A list of migrations, with each migration as a tuple of its
        index, name and function.
    """
    migrations_dir = os.path.abspath(os.path.dirname(__file__))

    migrations = {}

    for migration_file in os.listdir(migrations_dir):
        if not migration_file.endswith('.py'):
            continue
        if migration_file.startswith('_'):
            continue
        migration_name, _ = os.path.splitext(migration_file)
        migration_module = importlib.import_module('sampledb.models.migrations.' + migration_name)
        try:
            migration_index = migration_module.MIGRATION_INDEX
            migration_name = migration_module.MIGRATION_NAME
            migration_function = migration_module.run
        except AttributeError:
            continue
        # prevent duplicate migration indices
        assert migration_index not in migrations
        migrations[migration_index] = (migration_name, migration_function)

    migration_indices = list(sorted(migrations.keys()))

    # ensure no migration index is missing
    assert migration_indices == list(range(len(migration_indices)))

    sorted_migrations = []
    for index in migration_indices:
        name, function = migrations[index]
        sorted_migrations.append((index, name, function))
    return sorted_migrations



#!/usr/bin/env python3
# coding: utf-8

"""
This module defines a JSONEncoder and several data types which allow for de-/serialization to and from JSON using this
JSONEncoder. The format for serialized objects of these types is always a JavaScript object or Python dictionary,
containing a special attribute/key named '_type'. This value defines what class will be used for deserializing it, and
it also allows for some assumptions when dealing with the data, e.g. a "_type" equal to "quantity" guarantees that the
attribute/key "magnitude" will be exist and will be a Number.

While it would be simple to just verify this in Python, the advantage lies in using this information in a database,
where an invalid cast might lead to an erroneous query.

It has to be ensured, however, that no "_type" attribute or key can be set when not using these data types, as that
could potentially lead to the above assumptions and advantages being void. This can be verified when encoding by
setting JSONEncoder.STRICT = True, but the verification will greatly reduce the performance of the JSONEncoder.
"""

import datetime
import json
import pint

from .units import ureg

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


class _ReducingEncoder(json.JSONEncoder):
    """
    A JSONEncoder subclass that replaces all objects of serializable types with an empty dictionary.

    This class is used for the _contains_type(obj) function and should not be re-used for other purposes.
    """
    def default(self, obj):
        for type_name, cls in JSONEncoder.serializable_types.items():
            if isinstance(obj, cls):
                return {}
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)


def _contains_type(obj):
    """
    This function checks whether a JSON-serializable object does not contain any "_type" attributes or keys.

    :param obj: the object to check, should be created using the _ReducingEncoder
    :return: True if the object contains _type, False otherwise
    """
    if isinstance(obj, dict):
        return any(key == '_type' or _contains_type(value) for key, value in obj.items())
    if isinstance(obj, list):
        return any(_contains_type(value) for value in obj)
    return False


class JSONEncoder(json.JSONEncoder):
    """
    A JSONEncoder with support for registering additional JSON-serializable types.
    """

    STRICT = False
    serializable_types = {}

    def encode(self, obj):
        if JSONEncoder.STRICT:
            # Create a version of the object without any objects that are instances of the serializable types
            reduced_obj = json.loads(json.dumps(obj, cls=_ReducingEncoder))
            assert not _contains_type(reduced_obj)
        return super(JSONEncoder, self).encode(obj)

    def default(self, obj):
        for type_name, cls in JSONEncoder.serializable_types.items():
            if isinstance(obj, cls):
                obj = obj.to_json()
                obj['_type'] = type_name
                return obj
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)

    @classmethod
    def object_hook(cls, obj):
        """
        Function for json.loads() that will deserialize the objects serialized using default().

        :param obj: the dictionary to be deserialized
        :return: the deserialized object
        """
        if '_type' in obj:
            if obj['_type'] in cls.serializable_types:
                return cls.serializable_types[obj['_type']].from_json(obj)
        return obj

    @classmethod
    def serializable_type(cls, type_name):
        """
        Parameterized decorator that adds a class to the dictionary of serializable types using the given type name.
        """
        def serializable_type_decorator(c):
            cls.serializable_types[type_name] = c
            return c
        return serializable_type_decorator


@JSONEncoder.serializable_type('datetime')
class DateTime(object):
    JSON_SCHEMA = {
        'type': 'object',
        'properties': {
            '_type': {
                'enum': ['datetime']
            },
            'utc_datetime': {
                'type': 'string',
                'pattern': '^[0-9]{4}\\-[0-1][0-9]\\-[0-3][0-9]\\ [0-9]{2}:[0-9]{2}:[0-9]{2}$'
            }
        },
        'required': ['_type', 'utc_datetime'],
        'additionalProperties': False
    }

    FORMAT_STRING = '%Y-%m-%d %H:%M:%S'

    def __init__(self, utc_datetime=None):
        if utc_datetime is None:
            utc_datetime = datetime.datetime.utcnow()
        self.utc_datetime = utc_datetime.replace(microsecond=0)

    def __repr__(self):
        return '<{0}(utc_datetime={1})>'.format(type(self).__name__, self.utc_datetime.strftime(DateTime.FORMAT_STRING))

    def __eq__(self, other):
        return self.utc_datetime == other.utc_datetime

    def to_json(self):
        return {'utc_datetime': self.utc_datetime.strftime(DateTime.FORMAT_STRING)}

    @classmethod
    def from_json(cls, obj):
        return cls(datetime.datetime.strptime(obj['utc_datetime'], cls.FORMAT_STRING))


@JSONEncoder.serializable_type('quantity')
class Quantity(object):
    JSON_SCHEMA = {
        'type': 'object',
        'properties': {
            '_type': {
                'enum': ['quantity']
            },
            'dimensionality': {
                'type': 'string',
                'magnitude_in_base_units': 'string'
            },
            'magnitude_in_base_units': {
                'type': 'number'
            },
            'units': {
                'anyOf': [
                    {'type': 'null'},
                    {'type': 'string'}
                ]
            }
        },
        'required': ['_type', 'dimensionality', 'magnitude_in_base_units', 'units'],
        'additionalProperties': False
    }

    def __init__(self, magnitude, units):
        self.magnitude = float(magnitude)
        if units is None:
            self.units = None
            self.pint_units = ureg.Unit('1')
            self.magnitude_in_base_units = self.magnitude
        else:
            if isinstance(units, ureg.Unit):
                self.pint_units = units
                self.units = str(self.pint_units)
            else:
                self.units = units
                try:
                    self.pint_units = ureg.Unit(self.units)
                except (pint.errors.UndefinedUnitError, AttributeError):
                    raise ValueError("Invalid units '{}'".format(self.units))
            self.magnitude_in_base_units = ureg.Quantity(self.magnitude, self.pint_units).to_base_units().magnitude
        self.dimensionality = self.pint_units.dimensionality

    def __repr__(self):
        return '<{0}(magnitude={1.magnitude}, units="{1.units}")>'.format(type(self).__name__, self)

    def __eq__(self, other):
        return ureg.Quantity(self.magnitude, self.pint_units) == ureg.Quantity(other.magnitude, other.pint_units)

    def to_json(self):
        return {
            'magnitude_in_base_units': self.magnitude_in_base_units,
            'units': self.units,
            'dimensionality': str(self.dimensionality)
        }

    @classmethod
    def from_json(cls, obj):
        magnitude_in_base_units = obj['magnitude_in_base_units']
        units = obj['units']
        if units is None:
            pint_units = ureg.Unit('1')
        else:
            try:
                pint_units = ureg.Unit(units)
            except (pint.errors.UndefinedUnitError, AttributeError):
                raise ValueError("Invalid units '{}'".format(units))
        # convert magnitude back from base unit to desired unit
        pint_base_units = ureg.Quantity(1, pint_units).to_base_units().units
        magnitude = ureg.Quantity(magnitude_in_base_units, pint_base_units).to(pint_units).magnitude
        quantity = cls(magnitude, units)
        if pint_units.dimensionless:
            assert obj['dimensionality'] == 'dimensionless'
        else:
            assert (1 * pint_units).check(obj['dimensionality'])
        return quantity


@JSONEncoder.serializable_type('bool')
class Boolean(object):
    JSON_SCHEMA = {
        'type': 'object',
        'properties': {
            '_type': {
                'enum': ['bool']
            },
            'value': {
                'type': 'boolean'
            }
        },
        'required': ['_type', 'value'],
        'additionalProperties': False
    }

    def __init__(self, value):
        self.value = bool(value)

    def __repr__(self):
        return '<{0}(value={1.value})>'.format(type(self).__name__, self)

    def __eq__(self, other):
        return self.value == other.value

    def to_json(self):
        return {'value': self.value}

    @classmethod
    def from_json(cls, obj):
        return cls(obj['value'])


@JSONEncoder.serializable_type('text')
class Text(object):
    JSON_SCHEMA = {
        'type': 'object',
        'properties': {
            '_type': {
                'enum': ['text']
            },
            'text': {
                'type': 'string'
            }
        },
        'required': ['_type', 'text'],
        'additionalProperties': False
    }

    def __init__(self, text):
        self.text = text

    def __repr__(self):
        return '<{0}(text="{1.text}")>'.format(type(self).__name__, self)

    def __eq__(self, other):
        return self.text == other.text

    def to_json(self):
        return {'text': self.text}

    @classmethod
    def from_json(cls, obj):
        return cls(obj['text'])



# coding: utf-8
"""
There are several Markdown editors built into the SampleDB frontend, which
need to be able to upload images. This module allows uploading images which
are then assigned a random file name and can be used in the Markdown editors
using that name.
"""

import base64
import os

import flask
import flask_login

from . import frontend
from ..logic import markdown_images

_temporary_markdown_images = {}

IMAGE_FORMATS = {
    '.png': 'image/png',
    '.jpg': 'image/jpeg',
}


@frontend.route('/markdown_images/<file_name>')
@flask_login.login_required
def markdown_image(file_name):
    image_data = markdown_images.get_markdown_image(file_name, flask_login.current_user.id)
    if image_data is None:
        return flask.abort(404)
    file_extension = os.path.splitext(file_name)[1]
    return flask.Response(
        image_data,
        mimetype=IMAGE_FORMATS.get(file_extension, 'application/octet-stream')
    )


@frontend.route('/markdown_images/', methods=['POST'])
@flask_login.login_required
def upload_markdown_image():
    image_data_url = flask.request.get_data()
    for image_file_extension, image_content_type in IMAGE_FORMATS.items():
        image_data_url_prefix = b'data:' + image_content_type.encode('ascii') + b';base64,'
        if image_data_url.startswith(image_data_url_prefix):
            image_base64_data = image_data_url[len(image_data_url_prefix):]
            break
    else:
        return flask.abort(400)
    try:
        image_data = base64.b64decode(image_base64_data)
    except Exception:
        return flask.abort(400)
    file_name = markdown_images.store_temporary_markdown_image(image_data, image_file_extension, flask_login.current_user.id)
    return flask.url_for('.markdown_image', file_name=file_name)



# coding: utf-8
"""

"""

from . import authentication
from . import favorites
from . import files
from . import groups
from . import instruments
from . import instrument_log_entries
from . import locations
from . import markdown_images
from . import objects
from . import object_permissions
from . import projects
from . import settings
from . import users

from .actions import Action, ActionType
from .action_permissions import UserActionPermissions, GroupActionPermissions, ProjectActionPermissions, PublicActions
from .api_log import APILogEntry, HTTPMethod
from .authentication import Authentication, AuthenticationType
from .comments import Comment
from .favorites import FavoriteAction, FavoriteInstrument
from .files import File
from .groups import Group
from .instruments import Instrument
from .instrument_log_entries import InstrumentLogEntry
from .locations import Location, ObjectLocationAssignment
from .markdown_images import MarkdownImage
from .notifications import Notification, NotificationType, NotificationMode, NotificationModeForType
from .objects import Objects, Object
from .object_log import ObjectLogEntry, ObjectLogEntryType
from .object_permissions import UserObjectPermissions, GroupObjectPermissions, ProjectObjectPermissions, PublicObjects, DefaultUserPermissions, DefaultGroupPermissions, DefaultProjectPermissions, DefaultPublicPermissions
from .object_publications import ObjectPublication
from .permissions import Permissions
from .projects import Project, UserProjectPermissions, GroupProjectPermissions, SubprojectRelationship
from .settings import Settings
from .tags import Tag
from .users import User, UserType
from .user_log import UserLogEntry, UserLogEntryType


__all__ = [
    'api_log',
    'authentication',
    'favorites',
    'files',
    'groups',
    'instruments',
    'instrument_log_entries',
    'locations',
    'markdown_images',
    'objects',
    'object_permissions',
    'projects',
    'settings',
    'users',
    'Action',
    'ActionType',
    'UserActionPermissions',
    'GroupActionPermissions',
    'ProjectActionPermissions',
    'PublicActions',
    'APILogEntry',
    'Authentication',
    'AuthenticationType',
    'Comment',
    'FavoriteAction',
    'FavoriteInstrument',
    'File',
    'Group',
    'HTTPMethod',
    'Instrument',
    'InstrumentLogEntry',
    'Location',
    'ObjectLocationAssignment',
    'MarkdownImage',
    'Notification',
    'NotificationType',
    'NotificationMode',
    'NotificationModeForType',
    'Objects',
    'Object',
    'ObjectLogEntry',
    'ObjectLogEntryType',
    'ObjectPublication',
    'UserObjectPermissions',
    'GroupObjectPermissions',
    'ProjectObjectPermissions',
    'PublicObjects',
    'DefaultUserPermissions',
    'DefaultGroupPermissions',
    'DefaultProjectPermissions',
    'DefaultPublicPermissions',
    'Permissions',
    'Project',
    'UserProjectPermissions',
    'GroupProjectPermissions',
    'SubprojectRelationship',
    'Settings',
    'Tag',
    'User',
    'UserType',
    'UserLogEntry',
    'UserLogEntryType',
]



# coding: utf-8
"""

"""

from . import actions
from . import action_permissions
from . import api_log
from . import authentication
from . import comments
from . import datatypes
from . import errors
from . import export
from . import favorites
from . import files
from . import groups
from . import instruments
from . import instrument_log_entries
from . import ldap
from . import locations
from . import markdown_images
from . import notifications
from . import objects
from . import object_log
from . import object_relationships
from . import object_search
from . import object_permissions
from . import object_sorting
from . import projects
from . import publications
from . import rdf
from . import schemas
from . import security_tokens
from . import settings
from . import tags
from . import units
from . import users
from . import user_log
from . import utils
from . import where_filters

from ..models.objects import Objects

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


__all__ = [
    'actions',
    'action_permissions',
    'api_log',
    'authentication',
    'comments',
    'datatypes',
    'errors',
    'export',
    'favorites',
    'files',
    'groups',
    'instruments',
    'instrument_log_entries',
    'ldap',
    'locations',
    'markdown_images',
    'notifications',
    'objects',
    'object_log',
    'object_relationships',
    'object_search',
    'object_permissions',
    'object_sorting',
    'projects',
    'publications',
    'rdf',
    'security_tokens',
    'settings',
    'tags',
    'units',
    'users',
    'user_log',
    'utils',
    'where_filters',
]

Objects._data_validator = schemas.validate
Objects._schema_validator = schemas.validate_schema



# coding: utf-8
"""
Logic for users' favorite actions and instruments.
"""

from typing import List

from .users import get_user
from .actions import get_action
from .instruments import get_instrument
from ..models import FavoriteAction, FavoriteInstrument
from .. import db


def add_favorite_action(action_id: int, user_id: int) -> None:
    """
    Add an action to the list of a user's favorite actions.

    :param action_id: the ID of an existing action
    :param user_id: the ID of an existing user
    :raises UserDoesNotExistError: if the action does not exist
    :raises ActionDoesNotExistError: if the user does not exist
    """
    user = get_user(user_id)
    action = get_action(action_id)
    db.session.add(FavoriteAction(action_id=action.id, user_id=user.id))
    db.session.commit()


def remove_favorite_action(action_id: int, user_id: int) -> None:
    """
    Remove an action from the list of a user's favorite actions.

    If the action is not a favorite, this function does nothing.

    :param action_id: the ID of an existing action
    :param user_id: the ID of an existing user
    :raises UserDoesNotExistError: if the action does not exist
    :raises ActionDoesNotExistError: if the user does not exist
    """
    user = get_user(user_id)
    action = get_action(action_id)
    favorite_action = FavoriteAction.query.filter_by(action_id=action.id, user_id=user.id).first()
    if favorite_action is not None:
        db.session.delete(favorite_action)
        db.session.commit()


def get_user_favorite_action_ids(user_id: int) -> List[int]:
    """
    Get a list of a user's favorite actions.

    :param user_id: the ID of an existing user
    :returns: a list of IDs of favorite actions
    :raises UserDoesNotExistError: if the action does not exist
    """
    user = get_user(user_id)
    favorite_actions = FavoriteAction.query.filter_by(user_id=user.id).all()
    return [
        favorite_action.action_id
        for favorite_action in favorite_actions
    ]


def add_favorite_instrument(instrument_id: int, user_id: int) -> None:
    """
    Add an instrument to the list of a user's favorite instruments.

    :param instrument_id: the ID of an existing instrument
    :param user_id: the ID of an existing user
    :raises UserDoesNotExistError: if the instrument does not exist
    :raises InstrumentDoesNotExistError: if the user does not exist
    """
    user = get_user(user_id)
    instrument = get_instrument(instrument_id)
    db.session.add(FavoriteInstrument(instrument_id=instrument.id, user_id=user.id))
    db.session.commit()


def remove_favorite_instrument(instrument_id: int, user_id: int) -> None:
    """
    Remove an instrument from the list of a user's favorite instruments.

    If the instrument is not a favorite, this function does nothing.

    :param instrument_id: the ID of an existing instrument
    :param user_id: the ID of an existing user
    :raises UserDoesNotExistError: if the instrument does not exist
    :raises InstrumentDoesNotExistError: if the user does not exist
    """
    user = get_user(user_id)
    instrument = get_instrument(instrument_id)
    favorite_instrument = FavoriteInstrument.query.filter_by(instrument_id=instrument.id, user_id=user.id).first()
    if favorite_instrument is not None:
        db.session.delete(favorite_instrument)
        db.session.commit()


def get_user_favorite_instrument_ids(user_id: int) -> List[int]:
    """
    Get a list of a user's favorite instruments.

    :param user_id: the ID of an existing user
    :returns: a list of IDs of favorite instruments
    :raises UserDoesNotExistError: if the instrument does not exist
    """
    user = get_user(user_id)
    favorite_instruments = FavoriteInstrument.query.filter_by(user_id=user.id).all()
    return [
        favorite_instrument.instrument_id
        for favorite_instrument in favorite_instruments
    ]
